#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-agnostic type hint tester utilities** (i.e., callables
validating arbitrary objects to be type hints supported by :mod:`beartype`,
regardless of whether those hints comply with PEP standards or not).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.meta import URL_ISSUES
from beartype.roar import BeartypeDecorHintNonpepException
from beartype.typing import NoReturn
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import TypeException
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.nonpep.utilnonpeptest import (
    die_unless_hint_nonpep,
    is_hint_nonpep,
)
from beartype._util.hint.pep.utilpepget import get_hint_pep_typeargs_packed
from beartype._util.hint.pep.utilpeptest import (
    die_if_hint_pep_unsupported,
    is_hint_pep,
    is_hint_pep_supported,
)
from beartype._util.hint.pep.proposal.pep585 import (
    is_hint_pep585_builtin_subbed)
from beartype._util.hint.pep.proposal.pep484604 import is_hint_pep604

# ....................{ RAISERS                            }....................
def die_unless_hint(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed object is a **supported type hint**
    (i.e., object supported by the :func:`beartype.beartype` decorator as a
    valid type hint annotating callable parameters and return values).

    Specifically, this function raises an exception if this object is neither:

    * A **supported PEP-compliant type hint** (i.e., :mod:`beartype`-agnostic
      annotation compliant with annotation-centric PEPs currently supported
      by the :func:`beartype.beartype` decorator).
    * A **PEP-noncompliant type hint** (i.e., :mod:`beartype`-specific
      annotation intentionally *not* compliant with annotation-centric PEPs).

    Efficiency
    ----------
    This validator is effectively (but technically *not*) memoized. The passed
    ``exception_prefix`` parameter is usually unique to each call to this
    validator; memoizing this validator would uselessly consume excess space
    *without* improving time efficiency. Instead, this validator first calls the
    memoized :func:`.is_hint_pep` tester. If that tester returns :data:`True`,
    this validator immediately returns :data:`True` and is thus effectively
    memoized; else, this validator inefficiently raises a human-readable
    exception without memoization. Since efficiency is mostly irrelevant in
    exception handling, this validator remains effectively memoized.

    Parameters
    ----------
    hint : Hint
        Object to be validated.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    BeartypeDecorHintPepUnsupportedException
        If this object is a PEP-compliant type hint currently unsupported by
        the :func:`beartype.beartype` decorator.
    BeartypeDecorHintNonpepException
        If this object is neither a:

        * Supported PEP-compliant type hint.
        * Supported PEP-noncompliant type hint.
    '''

    # If this object is a supported type hint, reduce to a noop.
    if is_hint(hint):
        return
    # Else, this object is *NOT* a supported type hint. In this case,
    # subsequent logic raises an exception specific to the passed parameters.

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # BEGIN: Synchronize changes here with is_hint() below.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # If this hint is PEP-compliant *AND* currently unsupported by @beartype,
    # raise an exception.
    if is_hint_pep(hint):
        die_if_hint_pep_unsupported(
            hint=hint, exception_prefix=exception_prefix)
    # Else, this hint is PEP-noncompliant. In this case...

    # If this PEP-noncompliant hint is currently unsupported by @beartype, raise
    # an exception.
    die_unless_hint_nonpep(hint=hint, exception_prefix=exception_prefix)


def die_as_hint_unsupported(
    # Mandatory parameters.
    hint: object,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintNonpepException,
    exception_prefix: str = '',
) -> NoReturn:
    '''
    Unconditionally raise an exception describing the failure of the passed
    object to be a **supported type hint** (i.e., object supported by the
    :func:`beartype.beartype` decorator as a valid type hint annotating callable
    parameters and return values).

    This low-level raiser is intended to be called only after this object has
    already been validated to *not* be a supported type hint. This raiser
    centralizes similar logic previously duplicated throughout the codebase.

    Parameters
    ----------
    hint : object
        Object to be validated.
    exception_cls : type[Exception]
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintNonpepException`.
    exception_prefix : str, optional
        Human-readable substring prefixing the raised exception message.
        Defaults to the empty string.

    Raises
    ------
    exception_cls
        Unconditionally.
    '''
    assert isinstance(exception_cls, type), f'{repr(exception_cls)} not type.'
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')

    # Raise this generic exception message.
    raise exception_cls(
        f'{exception_prefix}type hint {repr(hint)} invalid or unrecognized. '
        f'This hint is either:\n'
        f'* PEP-noncompliant (and thus invalid). '
        f"This is your bad. Rejoice! @beartype isn't to blame for once.\n"
        f'* PEP-compliant but currently unsupported by @beartype '
        f'(and thus unrecognized). '
        f'This is our bad. Disaster! @beartype is to blame like always. '
        f'You suddenly feel encouraged to submit a feature request '
        f'for this unsupported hint to our friendly issue tracker at:\n'
        f'\t{URL_ISSUES}'
    )

# ....................{ TESTERS                            }....................
@callable_cached
def is_hint(hint: object) -> bool:
    '''
    :data:`True` only if the passed object is a **supported type hint** (i.e.,
    object supported by the :func:`beartype.beartype` decorator as a valid type
    hint annotating callable parameters and return values).

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : object
        Object to be validated.

    Returns
    -------
    bool
        :data:`True` only if this object is either:

        * A **PEP-compliant type hint** (i.e., :mod:`beartype`-agnostic
          annotation compliant with annotation-centric PEPs).
        * A **PEP-noncompliant type hint** (i.e., :mod:`beartype`-specific
          annotation intentionally *not* compliant with annotation-centric
          PEPs).

    Raises
    ------
    TypeError
        If this object is **unhashable** (i.e., *not* hashable by the builtin
        :func:`hash` function and thus unusable in hash-based containers like
        dictionaries and sets). All supported type hints are hashable.
    '''

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # BEGIN: Synchronize changes here with die_unless_hint() above.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Return true only if...
    return (
        # This is a PEP-compliant type hint supported by @beartype *OR*...
        is_hint_pep_supported(hint) if is_hint_pep(hint) else
        # This is a PEP-noncompliant type hint, which by definition is
        # necessarily supported by @beartype.
        is_hint_nonpep(hint=hint, is_forwardref_valid=True)
    )


#FIXME: Unit test us up, please.
def is_hint_cacheworthy(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed type hint is **cache-worthy.**

    A hint is cache-worthy if *all* of the following constraints hold:

    * This hint is *not* already **self-cached** (i.e., internally memoized) by
      its owner class or module. Although most hints *used* to self-cache, most
      modern hints no longer self-cache. Why? No idea. It's probably an
      accidental oversight in PEP implementations across different CPython
      developers that nobody ever bothered to report: e.g.,

      .. code-block:: pycon

         # PEP 585 type hints do *NOT* self-cache and are thus cache-worthy.
         >>> list[int] is list[int]
         False  # <-- horrible! pep 585, you are horrible!

         # PEP 484 type hints do self-cache and are thus *NOT* cache-worthy.
         >>> import typing
         >>> typing.List[int] is typing.List[int]
         True  # <-- wonderful. pep 484, you are wonderful.

    * The **machine-readable representation** (i.e., string returned by passing
      this hint to the :func:`repr` builtin) of this hint is **unambiguous**
      (i.e., there exists a one-to-one mapping between this representation and
      the contents of this representation such that only equal hints have equal
      representations). Although most hints have unambiguous representations,
      some hints have ambiguous representations -- including:

      * Type hints transitively parametrized by one or more **type variables**
        (i.e., :obj:`typing.TypeVar` objects) have ambiguous representations and
        are thus *not* cache-worthy. Note that the existence of even a single
        type variable parametrizing a single child type hint subscripting a
        parent type hint sadly renders the *entire* representation of that
        parent type hint ambiguous: e.g.,

        .. code-block:: pycon

           >>> from typing import TypeVar
           >>> T_int = TypeVar('T', bound=int)
           >>> T_str = TypeVar('T', bound=str)
           >>> repr(list[T_int])
           list[T]  # <-- ...ok
           >>> repr(list[T_str])
           list[T]  # <-- WUT U SAY!?
           >>> repr(list[T_int]) == repr(list[T_str])
           True  # <-- horrible! "TypeVar", you are horrible!

    Caveats
    -------
    This function *cannot* be meaningfully memoized, since the passed type hint
    is *not* guaranteed to be cached somewhere. Only functions passed cached
    type hints can be meaningfully memoized. Since this high-level function
    internally defers to unmemoized low-level functions that are ``O(n)`` for
    ``n`` the size of the inheritance hierarchy of this hint, this function
    should be called sparingly.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this type hint is uncached.

    See Also
    --------
    :func:`beartype._check.convert._convcoerce.coerce_hint_any`
        Further details.
    '''

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Avoid detecting the kind of this hint by calling any of the
    # get_hint_pep_sign_*() family of memoized getters. Doing so would consume
    # excess time and space when this hint is uncached, as passing this hint to
    # any of those getters would then cache against a hint that it is
    # functionally useless to cache against.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Return true only if this hint is...
    return (
        # *NOT* transitively parametrized by one or type variables (e.g.,
        # "list[tuple[T]]"). All first-party hints (i.e., hints defined by
        # CPython's standard library) and all third-party hints supported by
        # @beartype have unambiguous representations *EXCEPT* for hints
        # transitively parametrized by one or type variables, as type variables
        # themselves have ambiguous representations.
        not get_hint_pep_typeargs_packed(hint) and
        # Either...
        (
            # PEP 585-compliant (e.g., "list[str]"). This hint is *NOT*
            # self-caching (e.g., "list[str] is not list[str]").
            #
            # Note that this additionally includes all third-party type hints
            # that derive from the "types.GenericAlias" superclass, including:
            # * "numpy.typing.NDArray[...]" type hints.
            is_hint_pep585_builtin_subbed(hint) or
            # PEP 604-compliant (e.g., "int | str"). This hint is *NOT*
            # self-caching (e.g., "int | str is not int | str").
            #
            # Note that this hint could also be implicitly cached by coercing
            # this non-self-caching PEP 604-compliant union into a self-caching
            # PEP 484-compliant union (e.g., from "int | str" to "Union[int,
            # str]"). Since doing so would consume substantially more time for
            # *NO* tangible gain, we strongly prefer the current trivial and
            # efficient approach.
            is_hint_pep604(hint)
        )
    )

# ....................{ TESTERS ~ needs                    }....................
@callable_cached
def is_hint_needs_cls_stack(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed type hint is **type stack-dependent** (i.e.,
    if :mod:`beartype` requires the tuple of all classes lexically declaring the
    class variables or methods annotated by this hint to generate code
    type-checking this hint).

    This tester returns :data:`False` for most hints; only a small subset of
    hints are type stack-dependent. This includes:

    * :pep:`673`-compliant self type hint (i.e., :obj:`typing.Self`), which is
      contextually valid *only* inside a lexical class declaration.

    This tester is memoized for efficiency.

    Motivation
    ----------
    **This tester should only be called to decide whether memoized callables
    should be passed a type stack.** Passing memoized callables a type stack
    substantially reduces the likelihood of a cache hit and thus the
    average-case efficiency of calls to those callables. Notably:

    * Most type hints do *not* require a type stack.
    * Most type hints annotate class variables and methods of differing classes
      and thus do *not* share the same type stack.

    Ergo, passing memoized callables both a type hint *and* a type stack
    effectively unmemoizes those callables. Thankfully, callers can elide this
    inefficiency by calling this tester first; when this tester returns:

    * :data:`False`, the caller can safely pass memoized callables a type hint
      and :data:`None` for the type stack, thus preserving memoization. This is
      the common case and thus absolutely worth optimizing for.
    * :data:`True`, the caller has *no* choice but to pass memoized callables
      both a type hint *and* a type stack. Grr!

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this type hint is type stack-dependent.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.utilhintget import get_hint_repr

    # Machine-readable representation of this hint.
    hint_repr = get_hint_repr(hint)

    # Return true only if this representation embeds the representation of:
    # * A PEP 673-compliant self type hint (i.e., "typing.Self").
    #
    # Note:
    # * The fully-qualified names of exact typing modules (e.g., "typing",
    #   "typing_extensions") is intentionally ignored. Although we could
    #   certainly explicitly test for both, doing so would only needlessly
    #   reduce efficiency. By omitting explicit tests for both, this tester
    #   intentionally returns false positives for extremely edge case hints
    #   whose representations contain syntactically related but semantically
    #   unrelated substrings (e.g., "typing.Literal['.Self']", which is clearly
    #   *NOT* a PEP 673-compliant self type hint but erroneously matched as one
    #   by this heuristic). This is non-ideal but thankfully ignorable. See the
    #   "Motivation" section of the docstring for further commentary.
    # * This string is intentionally *NOT* preceded by a "." delimiter (e.g., as
    #   ".Self" rather than "Self"). Previously, this string was intentionally
    #   preceded by a "." delimiter; doing so satisfied most edge cases while
    #   reducing the likelihood of a false positive. Sadly, doing so also failed
    #   to match "Self" type hints stringified by PEP 563. Grrr!
    # * That the "in" operator is known to be the fastest means of performing
    #   substring matching in Python. Indeed:
    #   * "in" is faster than the str.find() method, which is *SUBSTANTIALLY*
    #     faster than the re.match() function, which is an entire of magnitude
    #     slower than "in".
    #   * re.match() only begins to catch up to "in" when concurrently testing
    #     for more than 10 or so substrings (e.g., r'(0|1|2|3|4|5|6|7|8|9)').
    #
    # See also the extensive timings documented at this StackOverflow question:
    #     https://stackoverflow.com/questions/4901523/whats-a-faster-operation-re-match-search-or-str-find
    return 'Self' in hint_repr



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`593` **type-checking expression snippets** (i.e.,
triple-quoted pure-Python string constants formatted and concatenated together
to dynamically generate boolean expressions type-checking arbitrary objects
against :pep:`593`-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import CallableStrFormat

# ....................{ CODE                               }....................
CODE_PEP593_VALIDATOR_PREFIX = '''('''
'''
:pep:`593`-compliant code snippet prefixing all code type-checking the current
pith against a :pep:`593`-compliant :obj:`typing.Annotated` type hint
subscripted by one or more :mod:`beartype.vale` validators.
'''


CODE_PEP593_VALIDATOR_SUFFIX = '''
{indent_curr})'''
'''
:pep:`593`-compliant code snippet suffixing all code type-checking the current
pith against each a :pep:`593`-compliant :class:`typing.Annotated` type hint
subscripted by one or more :mod:`beartype.vale` validators.
'''


CODE_PEP593_VALIDATOR_METAHINT = '''
{indent_curr}    {hint_child_placeholder} and'''
'''
:pep:`593`-compliant code snippet type-checking the current pith against the
**metahint** (i.e., first child type hint) subscripting a obj:`typing.Annotated`
type hint subscripted by one or more :mod:`beartype.vale` validators.
'''


CODE_PEP593_VALIDATOR_IS = '''
{indent_curr}    # True only if this pith satisfies this caller-defined
{indent_curr}    # validator of this annotated metahint.
{indent_curr}    {hint_child_expr} and'''
'''
:pep:`593`-compliant code snippet type-checking the current pith against
:mod:`beartype`-specific **data validator code** (i.e., caller-defined
:meth:`beartype.vale.BeartypeValidator._is_valid_code` string) of the current
child :mod:`beartype.vale` validator subscripting a parent :pep:`593`-compliant
:class:`typing.Annotated` type hint.

Caveats
-------
The caller is required to manually slice the trailing suffix ``" and"`` after
applying this snippet to the last subscripted argument of such a
:class:`typing.Annotated` type. While there exist alternate and more readable
means of accomplishing this, this approach is the optimally efficient.
'''

# ....................{ FORMATTERS                         }....................
# str.format() methods, globalized to avoid inefficient dot lookups elsewhere.
# This is an absurd micro-optimization. *fight me, github developer community*
CODE_PEP593_VALIDATOR_IS_format: CallableStrFormat = (
    CODE_PEP593_VALIDATOR_IS.format)
CODE_PEP593_VALIDATOR_METAHINT_format: CallableStrFormat = (
    CODE_PEP593_VALIDATOR_METAHINT.format)
CODE_PEP593_VALIDATOR_SUFFIX_format: CallableStrFormat = (
    CODE_PEP593_VALIDATOR_SUFFIX.format)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **typing module** utilities (i.e., callables dynamically testing
and importing attributes declared at module scope by either the standard
:mod:`typing` or third-party :mod:`typing_extensions` modules).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilModuleException
from beartype.typing import Any
from beartype._data.typing.datatyping import TypeException
from beartype._data.api.standard.datatyping import TYPING_MODULE_NAMES
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.module.utilmodimport import import_module_attr_or_none

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_typing_attr(
    # Mandatory parameters.
    typing_attr_basename: str,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilModuleException,
) -> bool:
    '''
    :data:`True` only if a **typing attribute** (i.e., object declared at module
    scope by either the :mod:`typing` or :mod:`typing_extensions` modules) with
    the passed unqualified name is importable from one or more of these
    modules.

    This function is effectively memoized for efficiency.

    Parameters
    ----------
    typing_attr_basename : str
        Unqualified name of the attribute to be imported from a typing module.
    exception_cls : Type[Exception]
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilModuleException`.

    Returns
    -------
    bool
        :data:`True` only if the :mod:`typing` or :mod:`typing_extensions`
        modules declare an attribute with this name.

    Raises
    ------
    exception_cls
        If this name is syntactically invalid.

    Warns
    -----
    BeartypeModuleUnimportableWarning
        If any of these modules raise module-scoped exceptions at importation
        time. That said, the :mod:`typing` and :mod:`typing_extensions` modules
        are scrupulously tested and thus unlikely to raise such exceptions.
    '''

    # Return true only if an attribute with this name is importable from either
    # the "typing" *OR* "typing_extensions" modules.
    #
    # Note that positional rather than keyword arguments are intentionally
    # passed to optimize memoization efficiency.
    return import_typing_attr_or_none(
        typing_attr_basename, exception_cls) is not None

# ....................{ IMPORTERS                          }....................
def import_typing_attr(
    # Mandatory parameters.
    typing_attr_basename: str,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilModuleException,
) -> Any:
    '''
    Dynamically import and return the **typing attribute** (i.e., object
    declared at module scope by either the :mod:`typing` or
    :mod:`typing_extensions` modules) with the passed unqualified name if
    importable from one or more of these modules *or* raise an exception
    otherwise (i.e., if this attribute is *not* importable from these modules).

    This function is effectively memoized for efficiency.

    Parameters
    ----------
    typing_attr_basename : str
        Unqualified name of the attribute to be imported from a typing module.
    exception_cls : Type[Exception]
        Type of exception to be raised by this function. Defaults to
        :class:`._BeartypeUtilModuleException`.

    Returns
    -------
    object
        Attribute with this name dynamically imported from a typing module.

    Raises
    ------
    exception_cls
        If either:

        * This name is syntactically invalid.
        * Neither the :mod:`typing` nor :mod:`typing_extensions` modules
          declare an attribute with this name.

    Warns
    -----
    BeartypeModuleUnimportableWarning
        If any of these modules raise module-scoped exceptions at importation
        time. That said, the :mod:`typing` and :mod:`typing_extensions` modules
        are scrupulously tested and thus unlikely to raise such exceptions.

    See Also
    --------
    :func:`beartype._util.module.utilmodimport.import_module_typing_any_attr_or_none`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.module.utilmodtest import is_module

    # Attribute with this name imported from either the "typing" or
    # "typing_extensions" modules if one or more of these modules declare this
    # attribute *OR* "None" otherwise.
    #
    # Note that positional rather than keyword arguments are intentionally
    # passed to optimize memoization efficiency.
    typing_attr = import_typing_attr_or_none(
        typing_attr_basename, exception_cls)

    # If none of these modules declare this attribute...
    if typing_attr is None:
        # Substrings prefixing and suffixing exception messages raised below.
        EXCEPTION_PREFIX = (
            f'Typing attributes "typing.{typing_attr_basename}" and '
            f'"typing_extensions.{typing_attr_basename}" not found. '
        )
        EXCEPTION_SUFFIX = (
            'We apologize for the inconvenience and hope you had a '
            'great dev cycle flying with Air Beartype, '
            '"Your Grizzled Pal in the Friendly Skies."'
        )

        # If the "typing_extensions" module is importable, raise an
        # appropriate exception.
        if is_module('typing_extensions'):
            raise exception_cls(
                f'{EXCEPTION_PREFIX} Please either '
                f'(A) update the "typing_extensions" package or '
                f'(B) update to a newer Python version. {EXCEPTION_SUFFIX}'
            )
        # Else, the "typing_extensions" module is unimportable. In this
        # case, raise an appropriate exception.
        else:
            raise exception_cls(
                f'{EXCEPTION_PREFIX} Please either '
                f'(A) install the "typing_extensions" package or '
                f'(B) update to a newer Python version. {EXCEPTION_SUFFIX}'
            )
    # Else, one or more of these modules declare this attribute.

    # Return this attribute.
    return typing_attr


#FIXME: Unit test us up, please.
def import_typing_attr_or_none(
    # Mandatory parameters.
    typing_attr_basename: str,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilModuleException,
) -> Any:
    '''
    Dynamically import and return the **typing attribute** (i.e., object
    declared at module scope by either the :mod:`typing` or
    :mod:`typing_extensions` modules) with the passed unqualified name if
    importable from one or more of these modules *or* :data:`None` otherwise
    otherwise (i.e., if this attribute is *not* importable from these modules).

    This function is effectively memoized for efficiency.

    Parameters
    ----------
    typing_attr_basename : str
        Unqualified name of the attribute to be imported from a typing module.
    exception_cls : Type[Exception]
        Type of exception to be raised by this function. Defaults to
        :class:`._BeartypeUtilModuleException`.

    Returns
    -------
    object
        Attribute with this name dynamically imported from a typing module.

    Raises
    ------
    exception_cls
        If this name is syntactically invalid.

    Warns
    -----
    BeartypeModuleUnimportableWarning
        If any of these modules raise module-scoped exceptions at importation
        time. That said, the :mod:`typing` and :mod:`typing_extensions` modules
        are scrupulously tested and thus unlikely to raise exceptions.

    See Also
    --------
    :func:`import_typing_attr_or_fallback`
        Further details.
    '''

    # One-liners in the rear view mirror may be closer than they appear.
    #
    # Note that parameters are intentionally passed positionally rather than by
    # keyword for memoization efficiency.
    return import_typing_attr_or_fallback(
        typing_attr_basename, None, exception_cls)


#FIXME: Unit test us up, please.
#FIXME: Leverage above, please.
@callable_cached
def import_typing_attr_or_fallback(
    # Mandatory parameters.
    typing_attr_basename: str,
    fallback: object,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilModuleException,
) -> Any:
    '''
    Dynamically import and return the **typing attribute** (i.e., object
    declared at module scope by either the :mod:`typing` or
    :mod:`typing_extensions` modules) with the passed unqualified name if
    importable from one or more of these modules *or* the passed fallback
    otherwise otherwise (i.e., if this attribute is *not* importable from these
    modules).

    Specifically, this function (in order):

    #. If the official :mod:`typing` module bundled with the active Python
       interpreter declares that attribute, dynamically imports and returns
       that attribute from that module.
    #. Else if the third-party (albeit quasi-official) :mod:`typing_extensions`
       module requiring external installation under the active Python
       interpreter declares that attribute, dynamically imports and returns
       that attribute from that module.
    #. Else, returns the passed fallback value.

    This function is memoized for efficiency.

    Parameters
    ----------
    typing_attr_basename : str
        Unqualified name of the attribute to be imported from a typing module.
    fallback : object
        Arbitrary value to be returned as a last-ditch fallback if *no* typing
        module declares this attribute.
    exception_cls : Type[Exception]
        Type of exception to be raised by this function. Defaults to
        :class:`._BeartypeUtilModuleException`.

    Returns
    -------
    object
        Attribute with this name dynamically imported from a typing module.

    Raises
    ------
    exception_cls
        If this name is syntactically invalid.

    Warns
    -----
    BeartypeModuleUnimportableWarning
        If any of these modules raise module-scoped exceptions at importation
        time. That said, the :mod:`typing` and :mod:`typing_extensions` modules
        are scrupulously tested and thus unlikely to raise exceptions.
    '''

    # Attribute with this name imported from the "typing" module if that module
    # declares this attribute *OR* "None" otherwise.
    typing_attr = import_module_attr_or_none(
        attr_name=f'typing.{typing_attr_basename}',
        exception_cls=exception_cls,
        exception_prefix='Typing attribute ',
    )

    # If the "typing" module does *NOT* declare this attribute...
    if typing_attr is None:
        # Attribute with this name imported from the "typing_extensions" module
        # if that module declares this attribute *OR* "None" otherwise.
        typing_attr = import_module_attr_or_none(
            attr_name=f'typing_extensions.{typing_attr_basename}',
            exception_cls=exception_cls,
            exception_prefix='Typing attribute ',
        )

        # If the "typing_extensions" module also does *NOT* declare this
        # attribute, fallback to the passed fallback value.
        if typing_attr is None:
            typing_attr = fallback
        # Else, the "typing_extensions" module declares this attribute.
    # Else, the "typing" module declares this attribute.

    # Return either this attribute if one or more of these modules declare this
    # attribute *OR* this fallback otherwise.
    return typing_attr

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
@callable_cached
def get_typing_attrs(typing_attr_basename: str) -> frozenset:
    '''
    Frozen set of all attributes with the passed unqualified basename declared
    by all importable typing modules, silently ignoring those modules failing to
    declare this attribute.

    This getter intentionally returns a set rather than a list. Why? Duplicates.
    The third-party :mod:`typing_extensions` module duplicates *all* type hint
    factories implemented by the standard :mod:`typing` module under the most
    recently released version of Python.

    This getter is memoized for efficiency.

    Attributes
    ----------
    typing_attr_basename : str
        Unqualified name of the attribute to be dynamically imported from
        each typing module.

    Yields
    ------
    set
        Set of all attributes with the passed unqualified basename declared by
        all importable typing modules.
    '''
    assert isinstance(typing_attr_basename, str), (
        f'{repr(typing_attr_basename)} not string.')

    # Set of all importable attributes to be returned by this getter.
    typing_attrs: set = set()

    # For the fully-qualified name of each quasi-standard typing module...
    for typing_module_name in TYPING_MODULE_NAMES:
        # Attribute with this name dynamically imported from that module if that
        # module defines this attribute *OR* "None" otherwise.
        typing_attr = import_module_attr_or_none(
            f'{typing_module_name}.{typing_attr_basename}')

        # If that module fails to define this attribute, silently continue to
        # the next module.
        if typing_attr is None:
            continue
        # Else, that module declares this attribute.

        # Append this attribute to this list.
        typing_attrs.add(typing_attr)

    # Return this set, coerced into a frozen set for caching purposes.
    return frozenset(typing_attrs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`612`-compliant :obj:`typing.Self` **unit tests**.

This submodule unit tests :pep:`612` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.10.0')
def test_decor_pep612() -> None:
    '''
    Test :pep:`612` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.10 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.door import (
        die_if_unbearable,
        is_bearable,
    )
    from beartype.roar import BeartypeDecorHintPep612Exception
    from beartype.typing import (
        ParamSpec,
        Tuple,
    )
    from beartype._data.typing.datatyping import DictStrToAny
    from pytest import raises

    # ....................{ LOCALs                         }....................
    # Arbitrary parameter specification.
    P = ParamSpec('P')

    # ....................{ CALLABLES                      }....................
    @beartype
    def as_shapes(
        *in_the_weird_clouds: P.args,
        **soft_mossy_lawns: P.kwargs,
    ) -> Tuple[tuple, DictStrToAny]:
        '''
        Arbitrary callable accepting both variadic positional and keyword
        parameters correctly annotated by :pep:`612`-compliant parameter
        specification positional and keyword parameter type hints, decorated by
        :func:`beartype.beartype`.
        '''

        # Return these variadic parameters as is.
        return in_the_weird_clouds, soft_mossy_lawns

    # ....................{ PASS                           }....................
    # Assert that a callable accepting both variadic positional and keyword
    # parameters correctly annotated by valid PEP 612-compliant parameter
    # specification positional and keyword parameter type hints returns the
    # expected object.
    assert as_shapes(
        'Beneath these canopies extend their swells',
        Fragrant_with_perfumed_herbs='and eyed with blooms',
    ) == (
        ('Beneath these canopies extend their swells',),
        {'Fragrant_with_perfumed_herbs': 'and eyed with blooms'},
    )

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating
    # callables accepting only one of rather than both variadic positional and
    # keyword parameters annotated by PEP 612-compliant parameter specification
    # keyword and positional parameter type hints.
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def their_noonday_watch(*and_sail_among_the_shades: P.args) -> None:
            pass
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def like_vaporous_shapes_half_seen(**beyond_a_well: P.kwargs) -> None:
            pass

    # Assert that @beartype raises the expected exception when decorating
    # callables accepting both variadic positional and keyword parameters such
    # that only one rather than both are annotated.
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def dark(*gleaming, **and_of_most_translucent_wave: P.kwargs) -> None:
            pass

    # Assert that @beartype raises the expected exception when decorating
    # callables accepting both variadic positional and keyword parameters such
    # that only one rather than both are annotated by PEP 612-compliant
    # parameter specification keyword and positional parameter type hints.
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def images_all(*the_woven_boughs: int, **above: P.kwargs) -> None:
            pass

    # Assert that @beartype raises the expected exception when decorating
    # callables accepting both variadic positional and keyword parameters
    # erroneously annotated by PEP 612-compliant parameter specification keyword
    # and positional parameter type hints (i.e., in erroneous order).
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def a_soul_dissolving_odour(
            *to_invite: P.kwargs,
            **to_some_more_lovely_mystery: P.args,
        ) -> None:
            pass

    # Assert that @beartype raises the expected exception when decorating
    # callables accepting non-variadic parameters erroneously annotated by PEP
    # 612-compliant parameter specification keyword and positional parameter
    # type hints.
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def through_the_dell(silence_and_twilight_here: P.args) -> None:
            pass
    with raises(BeartypeDecorHintPep612Exception):
        @beartype
        def twin_sisters(keep: P.kwargs) -> None:
            pass

    # Assert that statement-level runtime type-checkers raise the expected
    # exceptions when passed PEP 612-compliant parameter specification variadic
    # parameter type hints.
    with raises(BeartypeDecorHintPep612Exception):
        die_if_unbearable('Minute yet beautiful. One darkest glen', P.args)
    with raises(BeartypeDecorHintPep612Exception):
        is_bearable(
            'Sends from its woods of musk-rose, twined with jasmine,', P.kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`- or :pep:`604`-compliant **union type-checking code
factories** (i.e., low-level callables dynamically generating pure-Python code
snippets type-checking arbitrary objects against union type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Validate that PEP 695-compliant type aliases aliasing recursive unions
#behave as expected: e.g.,
#    type recursive_union = int | recursive_union
#FIXME: Likewise, note that our current *EXTREMELY* non-trivial handling of
#"hint_overrides"-based recursive unions below could probably benefit from being
#refactored into the same approach used to handle PEP 695-based recursive
#unions. The approach below is wild -- and not the good kind of "wild," either.

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    List,
    Tuple,
)
from beartype._check.metadata.hint.hintsmeta import HintsMeta
from beartype._check.metadata.hint.hintsane import (
    HINT_IGNORABLE,
    HintSane,
    DictHintSaneToAny,
    ListHintSane,
    TupleHintSane,
)
from beartype._data.code.datacodelen import LINE_RSTRIP_INDEX_OR
from beartype._data.code.pep.datacodepep484604 import (
    CODE_PEP484604_UNION_CHILD_PEP_format,
    CODE_PEP484604_UNION_CHILD_NONPEP_format,
    CODE_PEP484604_UNION_PREFIX,
    CODE_PEP484604_UNION_SUFFIX,
)
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import DictTypeToAny
from beartype._data.hint.sign.datahintsignset import HINT_SIGNS_UNION
from beartype._util.cache.pool.utilcachepoolinstance import (
    acquire_instance,
    release_instance,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.utilpepget import get_hint_pep_args
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.hint.pep.utilpeptest import is_hint_pep
from beartype._data.kind.datakindiota import SENTINEL

# ....................{ FACTORIES                          }....................
def make_hint_pep484604_check_expr(hints_meta: HintsMeta) -> None:
    '''
    Either a Python code snippet type-checking the current pith against the
    passed :pep:`484`- or :pep:`604`-compliant union type hint if this union is
    **unignorable** (i.e., subscripted by *no* ignorable child type hints) *or*
    :data:`None` otherwise (i.e., if this union is ignorable).

    This factory is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), due to accepting **context-sensitive
    parameters** (i.e., whose values contextually depend on context unique to
    the code being generated for the currently decorated callable) such as
    ``pith_curr_assign_expr``.

    Caveats
    -------
    Unions are non-physical abstractions of physical types and thus *not*
    themselves subject to type-checking; only the subscripted arguments of
    unions are type-checked. This differs from :mod:`typing` pseudo-containers
    like ``List[int]``, in which both the parent :obj:`typing.List` and child
    :class:`int` types represent physical types to be type-checked. Ergo, unions
    themselves impose no narrowing of the current pith expression and thus
    *cannot* by definition benefit from assignment expressions. This differs
    from :mod:`typing` pseudo-containers, which narrow the current pith
    expression and thus do benefit from assignment expressions.

    Parameters
    ----------
    hints_meta : HintsMeta
        Stack of metadata describing all visitable hints currently discovered by
        this breadth-first search (BFS).
    '''
    assert isinstance(hints_meta, HintsMeta), (
        f'{repr(hints_meta)} not "HintsMeta" object.')

    # ....................{ LOCALS                         }....................
    # Flattened tuple of two or more child hints subscripting this parent union
    # such that *ALL* nested child hints subscripting child unions are expanded
    # directly into this tuple, thus non-destructively eliminating child unions.
    #
    # Note that this getter is memoized and thus requires:
    # * Positional parameters.
    # * "hint_meta" instance variables to be explicitly passed rather than the
    #   "hint_meta" object in entirety. Why? Memoization, of course. Passing the
    #   "hint_meta" object in entirety would effectively inhibit the memoization
    #   of this getter, which entirely defeats the point.
    hint_childs_sane = _get_hint_pep484604_union_args_flattened(hints_meta)
    # print(f'Unflattened union {repr(hints_meta.hint_curr_meta.hint)}...')
    # print(f'...child hints {repr(hint_childs_sane)}.')

    # Dictionary whose keys comprise the set of all PEP-noncompliant child hints
    # subscripting this union and whose values are ignorable. Whereas sets fail
    # to preserve insertion order, dictionaries preserve insertion order. While
    # a set would be preferable for simplicity, the fact that sets are currently
    # unordered leaves us no choice but to abuse dictionaries.
    #
    # The order of union child hints is insignificant in the common case but
    # *NOT* the worst case, where order is significant. Examples include:
    # * A union over classes whose metaclasses define custom type-checking by
    #   overriding the __instancecheck__() dunder method. Callers might
    #   intentionally order those classes in that union to maximize
    #   type-checking efficiency. How? In descending order of the efficiency of
    #   those __instancecheck__() implementations. The most efficiently
    #   type-checkable classes would be listed first in that union; the least
    #   efficiently type-checkable classes, last.
    # * A union over classes registered as pseudo-subclasses with the
    #   collections.abc.ABCMeta.register() method. Callers might intentionally
    #   order those classes in that union to prevent type-checking code
    #   dynamically generated by beartype from raising unexpected exceptions in
    #   the event that those classes fail to fully satisfy the API of the ABC
    #   they were registered with. See also this issue:
    #       https://github.com/beartype/beartype/issues/499#issuecomment-2683400721
    #
    # Since PEP-compliant and -noncompliant child hints require fundamentally
    # different forms of type-checking, prefiltering child hints into these
    # dictionaries *BEFORE* generating code type-checking these child hints
    # improves both efficiency and maintainability.
    hint_childs_nonpep: DictTypeToAny = acquire_instance(dict)

    # Dictionary whose keys comprise the set of all PEP-compliant child hints
    # subscripting this union and whose values are ignorable. See above.
    hint_childs_sane_pep: DictHintSaneToAny = acquire_instance(dict)

    # ....................{ FILTER                         }....................
    #FIXME: Optimize by refactoring into a "while" loop. Naturally, profile that
    #doing so actually *IS* an optimization before doing so. *sigh*

    # For each child hint subscripting this union...
    for hint_child_sane in hint_childs_sane:
        #FIXME: Uncomment as desired for debugging. This test is currently a bit
        #too costly to warrant uncommenting.
        # Assert that this child hint is *NOT* shallowly ignorable. Why? Because
        # any union containing one or more shallowly ignorable child hints is
        # deeply ignorable and should thus have already been ignored after a
        # call to the is_hint_ignorable() tester passed this union on handling
        # the parent hint of this union.
        # assert (
        #     repr(hint_curr) not in HINTS_REPR_IGNORABLE_SHALLOW), (
        #     f'{hint_curr_exception_prefix} {repr(hint_curr)} child '
        #     f'{repr(hint_child)} ignorable but not ignored.')

        # Child hint encapsulated by this metadata.
        hint_child = hint_child_sane.hint

        #FIXME: *WOOOOOOOOOOOOOAH.* Waitjustaminute. Not all PEP-compliant type
        #hints are hashable! Seriously. Here's one obvious example:
        #    Annotated[object, []]
        #
        #There's absolutely *NO* way that's hashable. We... kinda messed up
        #here, folks. Thankfully, this edge case has yet to actually hit anyone.
        #This suggests that nearly all type hints of real-world interest are
        #hashable. Still, that's a pretty bad assumption. Let's generalize this
        #with "try" logic resembling:
        #    if is_hint_pep(hint_child):
        #        try:
        #            hint_childs_sane_pep[hint_child_sane] = None
        #        except TypeError:
        #            hint_childs_sane_pep_unhashable.append(hint_child_sane)
        #    else:
        #        hint_childs_nonpep[hint_child] = None  # type: ignore[index]
        #
        #As the above logic suggests, this edge case *ONLY* applies to
        #PEP-compliant type hints. All PEP-noncompliant type hints are
        #isinstanceable types -- which, by definition, are all hashable. \o/
        #
        #Naturally, we'll then need to:
        #* Rename "hint_childs_sane_pep" to "hint_childs_sane_pep_hashable".
        #* Define "hint_childs_sane_pep_unhashable" as a list above.
        #* Handle "hint_childs_sane_pep_unhashable" below. The most reliable way
        #  to do that is to:
        #  * First define a new tuple resembling:
        #      hint_childs_sane_pep_list = (
        #          list(hint_childs_sane_pep.keys()) +
        #          hint_childs_sane_pep_unhashable
        #      )
        #  * Then iterate over "hint_childs_sane_pep_list" rather than
        #    "hint_childs_sane_pep.keys()" below.

        # If this child hint is PEP-compliant, filter this child hint *AND* all
        # associated metadata (if any) into this dictionary of PEP-compliant
        # child hints.
        #
        # Note that this PEP-compliant child hint *CANNOT* also be filtered into
        # the dictionary of PEP-noncompliant child hints, even if this child
        # hint originates from a non-"typing" type (e.g., "List[int]" from
        # "list"). Why? Because that would then induce false positives when the
        # current pith shallowly satisfies this non-"typing" type but does *NOT*
        # deeply satisfy this child hint.
        if is_hint_pep(hint_child):
            hint_childs_sane_pep[hint_child_sane] = None
        # Else, this child hint is PEP-noncompliant. In this case, filter this
        # child hint into this dictionary of PEP-noncompliant child hints. Since
        # PEP-noncompliant hints are by definition associated with *NO*
        # meaningful metadata, silently ignore this metadata.
        else:
            hint_childs_nonpep[hint_child] = None  # type: ignore[index]

    # ....................{ FORMAT ~ non-pep               }....................
    # Initialize the code type-checking the current pith against these arguments
    # to the substring prefixing all such code.
    hints_meta.func_curr_code = CODE_PEP484604_UNION_PREFIX

    # If this union is subscripted by one or more PEP-noncompliant child hints,
    # generate and append efficient code type-checking these child hints
    # *BEFORE* less efficient code type-checking any PEP-compliant child hints
    # subscripting this union.
    if hint_childs_nonpep:
        hints_meta.func_curr_code += CODE_PEP484604_UNION_CHILD_NONPEP_format(
            # Python expression yielding the value of the current pith.
            # Specifically...
            pith_curr_expr=(
                # If this union is also subscripted by one or more
                # PEP-compliant child hints, prefer the expression assigning
                # this value to a local variable efficiently reused by
                # subsequent code generated for those PEP-compliant child
                # hints.
                hints_meta.pith_curr_assign_expr
                if hint_childs_sane_pep else
                # Else, this union is subscripted by *NO* PEP-compliant
                # child hints. Since this is the first and only test
                # generated for this union, prefer the expression yielding
                # the value of the current pith *WITHOUT* assigning this
                # value to a local variable, which would needlessly go
                # unused.
                hints_meta.pith_curr_expr
            ),
            # Python expression evaluating to a tuple of these arguments.
            #
            # Note that we would ideally avoid coercing this set into a
            # tuple when this set only contains one type by passing that
            # type directly to the _add_func_wrapper_local_type() function.
            # Sadly, the "set" class defines no convenient or efficient
            # means of retrieving the only item of a 1-set. Indeed, the most
            # efficient means of doing so is to iterate over that set and
            # immediately halt iteration:
            #     for first_item in muh_set: break
            #
            # While we *COULD* technically leverage that approach here,
            # doing so would also mandate adding multiple intermediate
            # tests, mitigating any performance gains. Ultimately, we avoid
            # doing so by falling back to the usual approach. See also this
            # relevant self-StackOverflow post:
            #       https://stackoverflow.com/a/40054478/2809027
            hint_curr_expr=hints_meta.add_func_scope_type_or_types(
                hint_childs_nonpep.keys()),
        )

    # ....................{ FORMAT ~ pep                   }....................
    # For the 0-based index of each PEP-compliant child hint of this union *AND*
    # that hint...
    for hint_child_sane_pep_index, hint_child_sane_pep in enumerate(
        hint_childs_sane_pep.keys()):
        # print(f'Enqueing union {hints_meta.hint_curr_meta.hint_sane.hint}...')
        # print(f'...PEP-compliant child {hint_child_sane_pep}.')

        # Code deeply type-checking this child hint.
        hints_meta.func_curr_code += CODE_PEP484604_UNION_CHILD_PEP_format(
            # Expression yielding the value of this pith.
            hint_child_placeholder=hints_meta.enqueue_hint_child_sane(
                hint_sane=hint_child_sane_pep,
                pith_expr=(
                    # If either...
                    #
                    # Then prefer the expression efficiently reusing the value
                    # previously assigned to a local variable by either the
                    # above conditional or prior iteration of the current
                    # conditional.
                    hints_meta.pith_curr_var_name
                    if (
                        # This union is also subscripted by one or more
                        # PEP-noncompliant child hints *OR*...
                        hint_childs_nonpep or
                        # This is any PEP-compliant child hint *EXCEPT* the
                        # first...
                        hint_child_sane_pep_index
                    ) else
                    # Then this union is not subscripted by any PEP-noncompliant
                    # child hints *AND* this is the first PEP-compliant child
                    # hint. In this case, preface this code with an expression
                    # assigning this value to a local variable efficiently
                    # reused by code generated by subsequent iteration.
                    #
                    # Note this child hint is guaranteed to be followed by at
                    # least one more child hint. Why? Because the "typing"
                    # module forces unions to be subscripted by two or more
                    # child hints. By deduction, those child hints *MUST* be
                    # PEP-compliant. Ergo, we need *NOT* explicitly validate
                    # that constraint here.
                    hints_meta.pith_curr_assign_expr
                ),
            ),
        )

    # ....................{ RETURN                         }....................
    # Release this pair of sets back to their respective pools.
    release_instance(hint_childs_nonpep)
    release_instance(hint_childs_sane_pep)

    # If this code is *NOT* its initial value, this union is subscripted by one
    # or more unignorable child hints and the above logic generated code
    # type-checking these child hints. In this case...
    if hints_meta.func_curr_code is not CODE_PEP484604_UNION_PREFIX:
        # Munge this code to...
        hints_meta.func_curr_code = (
            # Strip the erroneous " or" suffix appended by the last child hint
            # from this code.
            f'{hints_meta.func_curr_code[:LINE_RSTRIP_INDEX_OR]}'
            # Suffix this code by the substring suffixing all such code.
            f'{CODE_PEP484604_UNION_SUFFIX}'
        # Format the "indent_curr" prefix into this code, deferred above for
        # efficiency.
        ).format(indent_curr=hints_meta.indent_curr)
    # Else, this snippet is its initial value and thus ignorable.

# ....................{ PRIVATE ~ getters                  }....................
@callable_cached
def _get_hint_pep484604_union_args_flattened(
    hints_meta: HintsMeta) -> TupleHintSane:
    '''
    Flattened tuple of the two or more child hints subscripting the passed
    :pep:`604`- or :pep:`484`-compliant union hint such that *all* nested child
    hints subscripting *all* child union hints are **flattened** (i.e., expanded
    directly) into the returned tuple, thus non-destructively eliminating *all*
    child union hints of this parent union hint.

    This getter recursively flattens arbitrarily nested child union hints
    regardless of nesting depth in this parent union hint. Although non-trivial,
    doing so is required to support uncommon edge cases -- including:

    * :`pep:`695`-compliant type aliases, which technically enable users to
      deeply nest type aliases and thus unions to an arbitrarily deep nesting
      level: e.g.,

      .. code-block:: python

         # After reduction, the PEP 695-compliant "Level3" type alias defined
         # below effectively resembles:
         #     type Level3 = ((int | complex) | str) | float
         type Level1 = int | complex
         type Level2 = Level1 | str
         type Level3 = Level2 | float

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the only function calling this
    getter *is* itself memoized.

    Caveats
    -------
    **This getter exhibits time complexity** :math:`O(k)` **for k the total
    number of all transitive child hints of this union hint across all deeply
    nested child unions of this parent union hint.** A looser bound but somewhat
    more intelligible upper complexity bound would be :math:`O(n**m)`, where:

    * :m.th:`n` is the maximum number of child hints in any deeply nested child
      union of this parent union hint.
    * :math:`m` is the maximum depth of the most deeply nested child union of
      this parent union hint.

    Parameters
    ----------
    hints_meta : HintsMeta
        **Type hint type-checking metadata queue** (i.e., low-level fixed list
        of metadata describing all visitable type hints currently discovered by
        the breadth-first search (BFS) dynamically generating pure-Python
        type-checking code snippets in the
        :func:`beartype._check.code.codemain.make_check_expr` factory).

    Returns
    -------
    Tuple[HintSane, ...]
        Flattened tuple of the two or more child hints *or* **sanified child
        hint metadatum** (i.e., :class:`.HintSane` objects) subscripting
        this parent union hint.

    Raises
    ------
    BeartypeDecorHintPep604Exception
        If this tuple is empty.
    '''
    # print(f'[484/604] hint_curr_meta: {repr(hints_meta.hint_curr_meta)}')

    # ....................{ LOCALS                         }....................
    # Metadata encapsulating the currently visited PEP 484- or 604-compliant
    # union type hint as well as that hint.
    union_hint_sane = hints_meta.hint_curr_meta.hint_sane
    union_hint = union_hint_sane.hint

    # ....................{ LOCALS ~ child                 }....................
    # Tuple of the two or more child hints subscripting this union.
    hint_childs = get_hint_pep_args(union_hint)

    # Number of these child hints.
    hint_childs_len = len(hint_childs)

    # Assert this union to be subscripted by two or more child hints.
    #
    # Note this should *ALWAYS* be the case, as:
    # * The unsubscripted "typing.Union" type hint factory is explicitly listed
    #   in the "HINTS_REPR_IGNORABLE_SHALLOW" set and should thus have already
    #   been ignored when present.
    # * The "typing" module explicitly prohibits empty union subscription: e.g.,
    #       >>> typing.Union[]
    #       SyntaxError: invalid syntax
    #       >>> typing.Union[()]
    #       TypeError: Cannot take a Union of no types.
    # * The "typing" module reduces unions of one child hint to that hint: e.g.,
    #     >>> import typing
    #     >>> typing.Union[int]
    #     int
    assert hint_childs_len >= 2, (
        f'{hints_meta.exception_prefix}'
        f'PEP 484 or 604 union type hint {repr(union_hint)} either '
        f'unsubscripted or subscripted by only one child type hint.'
    )

    # ....................{ LOCALS ~ list                  }....................
    # Input stack of all currently unflattened transitive child hints of this
    # union to be visited by the depth-first search (DFS) below such that each
    # item of this stack is a 2-tuple (hint_child_insane, hint_parent_sane),
    # where:
    # * "hint_child_insane" is an transitive child hint of this union that has
    #   yet to be sanified and thus possibly expanded into a nested union hint
    #   requiring flattening into this root union hint.
    # * "hint_parent_sane" is sanified metadata encapsulating the direct parent
    #   hint of "hint_child_insane". In theory, this should *ALWAYS* be a union
    #   hint -- either this root union hint *OR* a nested union hint thereof.
    hint_childs_insane_unflattened: List[Tuple[Hint, HintSane]] = (
        acquire_instance(list))

    # Efficiently initialize this input stack to the non-empty list of all
    # 2-tuples (hint_child_insane, union_hint_sane) containing all direct child
    # hints of this union, equivalent to the following inefficient iteration:
    #     for hint_child in hint_childs:
    #         hint_childs_insane_unflattened.append((
    #             hint_child, union_hint_sane))
    #
    # Specifically:
    # * "(union_hint_sane,)*hint_childs_len" expands to the n-tuple containing
    #   "n" repetitions of this root union hint, where n is the number of child
    #   hints subscripting this root union hint.
    # * This call to the zip() builtin creates an iterable of 2-tuples
    #   (hint_child_insane, union_hint_sane) for each such child hint.
    hint_childs_insane_unflattened.extend(zip(
        hint_childs, (union_hint_sane,)*hint_childs_len))

    # Output stack of all previously flattened transitive child hints of this
    # union that have already been visited by this DFS such that each item of
    # this list is sanified metadata encapsulating each such child hint,
    # initialized to the empty list. Equivalently, this is the list of all
    # sanified child hints from which to reconstitute this union below.
    #
    # Note that this stack orders these child hints in the *REVERSE* order that
    # these child hints were originally ordered by the user in this union.
    hint_childs_sane_flattened: ListHintSane = acquire_instance(list)

    # ....................{ SEARCH                         }....................
    # Repeatedly flatten all currently unflattened transitive child hints of
    # this union into an increasingly flat union until this union can no longer
    # be flattened (i.e., after these child hints have all been flattened).
    # Recursively perform a depth-first search (DFS) over these child hints with
    # non-recursive iteration for both efficiency and simplicity. While
    # non-recursive iteration is often non-trivial, this implementation is
    # surprisingly trivial.
    #
    # While the input stack of these child hints is still non-empty, one or more
    # transitive child hints of this union have yet to be flattened. Then...
    while hint_childs_insane_unflattened:
        # ....................{ SANIFY                     }....................
        # Currently unflattened transitive child hint to be flattened defined as
        # the 2-tuple (hint_child_insane, hint_parent_sane), where:
        # * "hint_child_insane" is an transitive child hint of this union that
        #   has yet to be sanified and thus possibly expanded into a nested
        #   union hint requiring flattening into this root union hint.
        # * "hint_parent_sane" is sanified metadata encapsulating the direct
        #   parent hint of "hint_child_insane". In theory, this should *ALWAYS*
        #   be a union hint -- either this root union hint *OR* a nested union
        #   hint sanified by a prior iteration of this loop below.
        hint_child_insane, hint_parent_sane = (
            hint_childs_insane_unflattened.pop())

        # Metadata encapsulating the sanification of the currently unflattened
        # transitive child hint to be flattened.
        hint_child_sane: HintSane = SENTINEL  # type: ignore[assignment]

        # Sanified hint metadata encapsulating the sanification of this
        # possibly insane child hint with respect to the previously sanified
        # metadata encapsulating the direct parent hint of this child hint.
        #
        # Note that this sanification is intentionally performed *BEFORE* the
        # sign of this child hint is tested. Why? Because some reductions expand
        # an arbitrary hint into a union. This includes:
        # * PEP 695-compliant type aliases aliased to unions. See above!
        # * The PEP-noncompliant "float' and "complex" types, implicitly
        #   expanded to the PEP 484-compliant "float | int" and "complex | float
        #   | int" type hints (respectively) when the non-default
        #   "conf.is_pep484_tower=True" parameter is enabled.
        # * User-defined "hint_overrides", a generalization of the prior item.
        #
        # Likewise, note that this sanification implicitly handles *ALL*
        # recursive edge cases. We need *NOT* do so explicitly above or below.
        # Notably, if this child hint has already been sanified by a previously
        # performed sanification in this recursive tree of all previously
        # performed sanifications, there now exist two divergent edge cases.
        # Either:
        # * This child hint has intrinsic value and *MUST* thus be preserved as
        #   a member of this union. This edge case currently arises *ONLY*
        #   through hint overrides. Consider overriding the child hint "float"
        #   with the PEP 604-compliant union "int | float". In this case:
        #   * The initial sanification of the child hint "float" performed above
        #     will expand that hint to "int | float".
        #   * The unflattening performed below will then append the child hints
        #     "int" and "float" for subsequent iteration by this "while" loop.
        #   * The child hint "float" will then be detected as having been
        #     previously sanified. This child hint *CANNOT* be removed or
        #     ignored, as doing so would omit this child hint from this union.
        #     This child hint thus has intrinsic value and *MUST* thus be
        #     preserved as a member of this union.
        #
        #   This edge case is implicitly handled by the
        #   hints_meta.sanify_hint_child() method called below, which
        #   transitively calls the reduce_hint() function, which detects
        #   recursion in an overridden hint and implicitly returns that hint as
        #   is *WITHOUT* recursively overriding that hint yet again.
        # * This child hint lacks intrinsic value and *MUST* thus be removed as
        #   a member of this union. This edge case currently arises *ONLY*
        #   through PEP 695-compliant recursive type aliases. Consider:
        #       type RecursiveUnion = int | RecursiveUnion
        #
        #   The recursive type alias "RecursiveAlias" merely has symbolic value
        #   and thus lacks intrinsic value. This type alias is semantically
        #   equivalent to the non-recursive type alias:
        #       type NonrecursiveAlias = int
        #
        #   Similar logic as with the prior edge case then applies, except that
        #   this hint *SHOULD* be removed and ignored (rather than preserved).
        #   Thankfully, this edge case as well is implicitly handled by the
        #   hints_meta.sanify_hint_child() method called below, which
        #   transitively calls the reduce_hint() function, which detects
        #   recursion in a non-overridden hint and explicitly returns
        #   "HINT_SANE_IGNORABLE" rather than recursing infinitely into that hint.
        # print(f'Sanifying union child hint {repr(hint_child)} under {repr(conf)}...')
        hint_child_sane = hints_meta.sanify_hint_child(
            hint_child_insane=hint_child_insane,
            hint_parent_sane=hint_parent_sane,
        )
        # print(f'Sanified union child hint {hint_child_insane} to {hint_child_sane}!')

        # Assert this child hint to be unignorable. The previously applied
        # reduction for PEP 484- and 604-compliant union hints (i.e., the
        # reduce_hint_pep484604() reducer) *SHOULD* have already ignored union
        # hints containing *ANY* ignorable child hints. However, this union was
        # *NOT* ignored! By elimination, this union must contain *NO* ignorable
        # child hints.
        assert hint_child_sane.hint is not HINT_IGNORABLE, (
            f'Union {repr(union_hint)} '
            f'containing ignorable child {repr(hint_child_insane)} '
            f'not itself ignored (i.e., reduced to "HINT_IGNORABLE" singleton).'
        )

        # ....................{ UNION                      }....................
        # Sane child hint encapsulated by this metadata.
        hint_child = hint_child_sane.hint

        # Sign of this sanified child hint if this hint is PEP-compliant *OR*
        # "None" otherwise (i.e., if this hint is PEP-noncompliant).
        hint_child_sign = get_hint_pep_sign_or_none(hint_child)

        # If this child hint is itself a child union nested in this parent
        # union, explicitly flatten this nested union by appending *ALL* child
        # child hints subscripting this child union onto this parent union.
        #
        # Note that this edge case currently *ONLY* arises when this child hint
        # has been expanded by the above call to the sanify_hint_child()
        # function from a non-union (e.g., "float") into a union (e.g., "float |
        # int"). The standard PEP 484-compliant "typing.Union" type hint factory
        # already implicitly flattens nested unions: e.g.,
        #     >>> from typing import Union
        #     >>> Union[float, Union[int, str]]
        #     typing.Union[float, int, str]
        if hint_child_sign in HINT_SIGNS_UNION:
            # Tuple of all child child hints subscripting this child union.
            hint_child_childs = get_hint_pep_args(hint_child)
            # print(f'Expanding union {repr(hint_curr)} with child union {repr(hint_child_childs)}...')

            # For each child child type subscripting this child union,
            # efficiently append the 2-tuple containing both this unsanified
            # child child hint *AND* this sanified child hint metadata to this
            # stack. This is equivalent to the following inefficient iteration:
            #     for hint_child_child in hint_child_childs:
            #         hint_childs_insane_unflattened.append((
            #             hint_child_child, hint_child_sane))
            #
            # See above for further discussion dissecting this madness.
            hint_childs_insane_unflattened.extend(zip(
                hint_child_childs, (hint_child_sane,)*len(hint_child_childs)))
        # ....................{ NON-UNION                  }....................
        # Else, this child hint is *NOT* itself a union. In this case, append
        # this child hint to this parent union.
        else:
            hint_childs_sane_flattened.append(hint_child_sane)
            # print(f'Flattened union child {repr(hint_child)} to {repr(hint_child_sane)}...')

    # ....................{ RETURN                         }....................
    # Tuple of all flattened transitive child hints of this union, reversed so
    # as to undo the reverse ordering applied by the above algorithm and thus
    # preserve the original ordering of these child hints of this union.
    hint_childs_sane_flattened_tuple = tuple(reversed(
        hint_childs_sane_flattened))

    # Release these stacks back to their respective pool.
    release_instance(hint_childs_insane_unflattened)
    release_instance(hint_childs_sane_flattened)
    # print(f'Flattened union to {repr(hint_childs_sane)}...')

    # Return this tuple.
    return hint_childs_sane_flattened_tuple



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **hint sign logic mappings** (i.e., dictionary globals mapping from
signs uniquely identifying various kinds of type hints to corresponding
dataclasses encapsulating all low-level Python code snippets and associated
metadata required to dynamically generate high-level Python code snippets fully
type-checking those type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Dict
from beartype._check.logic.logcls import (
    HintLogicABC,
    HintLogicQuasiiterable,
    HintLogicReiterable,
    HintLogicSequence,
)
from beartype._data.hint.sign.datahintsigncls import HintSign

# ....................{ MAPPINGS                           }....................
# Initialized by the _init() function defined below.
HINT_SIGN_PEP484585_CONTAINER_TO_LOGIC: Dict[HintSign, HintLogicABC] = {}
'''
Dictionary mapping from the sign uniquely identifying each applicable kind of
**standard single-argument container type hint** (i.e., :pep:`484`- or
:pep:`585`-compliant type hint describing a standard container, subscripted by
exactly one child type hint constraining *all* items contained in that
container) to the hint sign logic dataclass dynamically generating Python code
snippets type-checking that kind of type hint.
'''

HINT_SIGN_PEP484585_CONTAINER_TO_LOGIC_get = (
    HINT_SIGN_PEP484585_CONTAINER_TO_LOGIC.get)
'''
:meth:`dict.get` method bound to the
:data:`.HINT_PEP484585_CONTAINER_TO_LOGIC` global for efficiency.
'''

# ..................{ PRIVATE ~ main                         }..................
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer function-specific imports.
    from beartype._data.hint.sign.datahintsignset import (
        HINT_SIGNS_QUASIITERABLE,
        HINT_SIGNS_REITERABLE,
        HINT_SIGNS_SEQUENCE,
    )

    # ....................{ DEFINE                         }....................
    # Hint sign logic singletons.
    hint_logic_quasiiterable = HintLogicQuasiiterable()
    hint_logic_reiterable = HintLogicReiterable()
    hint_logic_sequence = HintLogicSequence()

    # For each sign identifying a single-argument uniiterable hint...
    for hint_sign in HINT_SIGNS_QUASIITERABLE:
        # Map this sign to this logic dataclass.
        HINT_SIGN_PEP484585_CONTAINER_TO_LOGIC[hint_sign] = (
            hint_logic_quasiiterable)

    # For each sign identifying a single-argument reiterable hint...
    for hint_sign in HINT_SIGNS_REITERABLE:
        # Map this sign to this logic dataclass.
        HINT_SIGN_PEP484585_CONTAINER_TO_LOGIC[hint_sign] = (
            hint_logic_reiterable)

    # For each sign identifying a single-argument sequence hint...
    for hint_sign in HINT_SIGNS_SEQUENCE:
        # Map this sign to this logic dataclass.
        HINT_SIGN_PEP484585_CONTAINER_TO_LOGIC[hint_sign] = hint_logic_sequence


# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) factories** (i.e., low-level callables
creating and returning various types of nodes, typically for inclusion in the
currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Attribute,
    Call,
    Constant,
    Expr,
    FormattedValue,
    ImportFrom,
    Module,
    Name,
    alias,
    expr,
    keyword,
    parse as ast_parse,
)
from beartype.roar import BeartypeClawImportAstException
from beartype.roar._roarexc import _BeartypeUtilAstException
from beartype.typing import (
    List,
    Optional,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._data.api.standard.dataast import (
    NODE_CONTEXT_LOAD,
    NODE_CONTEXT_STORE,
)
# from beartype._data.typing.datatyping import NodesList
from beartype._data.kind.datakindsequence import LIST_EMPTY
from beartype._util.ast.utilastmunge import copy_node_metadata

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please. When we do, remove the "pragma: no cover" from
#the body of this getter below.
def make_node_from_code_snippet(code_snippet: str) -> AST:
    '''
    Abstract syntax tree (AST) node parsed from the passed (presumably)
    triple-quoted string defining a single child object.

    This function is principally intended to be called from our test suite as a
    convenient means of "parsing" triple-quoted strings into AST nodes.

    Caveats
    -------
    **This function assumes that this string defines only a single child
    object.** If this string defines either no *or* two or more child objects,
    an exception is raised.

    Parameters
    ----------
    code_snippet : str
        Triple-quoted string defining a single child object.

    Returns
    -------
    AST
        AST node encapsulating the object defined by this string.

    Raises
    -------
    _BeartypeUtilAstException
        If this string defines either no *or* two or more child objects.
    '''
    assert isinstance(code_snippet, str), f'{repr(code_snippet)} not string.'

    # "ast.Module" AST tree parsed from this string.
    node_module = ast_parse(code_snippet)

    # If this node is *NOT* actually a module node, raise an exception.
    if not isinstance(node_module, Module):  # pragma: no cover
        raise _BeartypeUtilAstException(
            f'{repr(node_module)} not AST module node.')
    # Else, this node is a module node.

    # List of all direct child nodes of this parent module name.
    nodes_child = node_module.body

    # If this module node contains either no *OR* two or more child nodes, raise
    # an exception.
    if len(nodes_child) != 1:  # pragma: no cover
        raise _BeartypeUtilAstException(
            f'Python code {repr(code_snippet)} defines '
            f'{len(nodes_child)} != 1 child objects.'
        )
    # Else, this module node contains exactly one child node.

    # Return this child node.
    return nodes_child[0]

# ....................{ FACTORIES ~ attribute              }....................
#FIXME: Unit test us up, please.
def make_node_object_attr_load(
    # Mandatory parameters.
    attr_name: str,
    node_sibling: AST,

    # Optional parameters.
    node_obj: Optional[AST] = None,
    obj_name: Optional[str] = None,
) -> Attribute:
    '''
    Create and return a new **object attribute access abstract syntax tree (AST)
    node** (i.e., node encapsulating an access of an object attribute) of the
    passed object with the passed attribute name.

    Note that exactly one of the ``node_obj`` and ``obj_name`` parameters *must*
    be passed. If neither or both of these parameters are passed, an exception
    is raised.

    Parameters
    ----------
    attr_name : str
        Unqualified basename of the attribute of this object to be accessed.
    node_sibling : AST
        Sibling node to copy source code metadata from.
    node_obj : Optional[expr]
        Either:

        * If the caller prefers supplying the name node accessing the parent
          object to load this attribute from, that node.
        * Else, :data:`None`. In this case, the caller *must* pass the
          ``obj_name`` parameter.

        Defaults to :data:`None`.
    obj_name : Optional[str]
        Either:

        * If the caller prefers supplying the unqualified basename of the parent
          object to load this attribute from in the current lexical scope,
          that basename.
        * Else, :data:`None`. In this case, the caller *must* pass the
          ``node_obj`` parameter.

        Defaults to :data:`None`.

    Returns
    -------
    Attribute
        Object attribute node accessing this attribute of this object.

    Raises
    ------
    BeartypeClawImportAstException
        If either:

        * Neither the ``node_obj`` nor ``obj_name`` parameters are passed.
        * Both of the ``node_obj`` and ``obj_name`` parameters are passed.
    '''
    assert isinstance(attr_name, str), f'{repr(attr_name)} not string.'

    # If the caller passed *NO* name node accessing the parent object to load
    # this attribute from...
    if not node_obj:
        # If the caller also passed *NO* unqualified basename of that object,
        # raise an exception.
        if not obj_name:
            raise BeartypeClawImportAstException(
                f'Attribute "{attr_name}" parent object undefined '
                f'(i.e., neither "node_obj" nor "obj_name" parameters passed).'
            )
        # Else, the caller also passed the unqualified basename of that object.

        # Child node accessing that object with this basename.
        node_obj = make_node_name_load(name=obj_name, node_sibling=node_sibling)
    # Else, the caller passed a name node accessing that object.
    #
    # If the caller also passed the unqualified basename of that object, raise
    # an exception.
    elif obj_name:
        raise BeartypeClawImportAstException(
            f'Attribute "{attr_name}" parent object overly defined '
            f'(i.e., both "node_obj" and "obj_name" parameters passed).'
        )
    # Else, the caller passed *NO* unqualified basename of that object.
    #
    # In any case, the "node_obj" variable is now the desired object node.
    assert isinstance(node_obj, expr), (
        f'{repr(node_obj)} not AST expression node.')

    # Object attribute node accessing this attribute of this object.
    node_attribute_load = Attribute(
        value=node_obj, attr=attr_name, ctx=NODE_CONTEXT_LOAD)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_attribute_load)

    # Return this node.
    return node_attribute_load

# ....................{ FACTORIES ~ attribute : name       }....................
#FIXME: Unit test us up.
def make_node_name_load(name: str, node_sibling: AST) -> Name:
    '''
    Create and return a new **attribute access abstract syntax tree (AST) node**
    (i.e., node encapsulating an access of an attribute) in the current lexical
    scope with the passed name.

    Parameters
    ----------
    name : str
        Fully-qualified name of the attribute to be accessed.
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    -------
    Name
        Name node accessing this attribute in the current lexical scope.
    '''
    assert isinstance(name, str), f'{repr(name)} not string.'

    # Child node accessing this attribute in the current lexical scope.
    node_name = Name(name, ctx=NODE_CONTEXT_LOAD)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_name)

    # Return this child node.
    return node_name


#FIXME: Unit test us up.
def make_node_name_store(name: str, node_sibling: AST) -> Name:
    '''
    Create and return a new **attribute assignment abstract syntax tree (AST)
    node** (i.e., node encapsulating an assignment of an attribute) in the
    current lexical scope with the passed name.

    Parameters
    ----------
    name : str
        Fully-qualified name of the attribute to be assigned.
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    -------
    Name
        Name node assigning this attribute in the current lexical scope.
    '''
    assert isinstance(name, str), f'{repr(name)} not string.'

    # Child node assigning this attribute in the current lexical scope.
    node_name = Name(name, ctx=NODE_CONTEXT_STORE)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_name)

    # Return this child node.
    return node_name

# ....................{ FACTORIES ~ call                   }....................
#FIXME: Unit test us up, please.
def make_node_call_expr(*args, node_sibling: AST, **kwargs) -> Expr:
    '''
    Create and return a new **callable call expression abstract syntax tree
    (AST) node** (i.e., node encapsulating a Python expression expressing a call
    to an arbitrary function or method) calling the function or method with the
    passed name, positional arguments, and keyword arguments.

    Parameters
    ----------
    node_sibling : AST
        Sibling node to copy source code metadata from.

    All remaining passed positional and keyword parameters are passed to the
    lower-level :func:`.make_node_call` factory function as is.

    Returns
    -------
    Expr
        Expression node calling this callable with these parameters.
    '''

    # Child node calling this callable.
    node_func_call = make_node_call(*args, node_sibling=node_sibling, **kwargs)  # type: ignore[misc]

    # Child node expressing this call as a Python expression.
    node_func = Expr(node_func_call)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_func)

    # Return this expression node.
    return node_func


#FIXME: Unit test us up, please.
def make_node_call(
    # Mandatory parameters.
    func_name: str,
    node_sibling: AST,

    # Optional parameters.
    nodes_args: List[expr] = LIST_EMPTY,
    nodes_kwargs: List[keyword] = LIST_EMPTY,
) -> Call:
    '''
    Create and return a new **callable call abstract syntax tree (AST) node**
    (i.e., node encapsulating a call to an arbitrary function or method)
    calling the function or method with the passed name, positional arguments,
    and keyword arguments.

    Parameters
    ----------
    func_name : str
        Fully-qualified name of the module to import this attribute from.
    node_sibling : AST
        Sibling node to copy source code metadata from.
    nodes_args : List[expr], optional
        List of zero or more **positional parameter AST expression nodes**
        comprising the tuple of all positional parameters to be passed to this
        call. Defaults to the empty list.
    nodes_kwargs : List[keyword], optional
        List of zero or more **keyword parameter AST nodes** comprising the
        dictionary of all keyword parameters to be passed to this call. Defaults
        to the empty list.

    Returns
    -------
    Call
        Callable call node calling this callable with these parameters.
    '''
    assert isinstance(nodes_args, list), f'{repr(nodes_args)} not list.'
    assert isinstance(nodes_kwargs, list), f'{repr(nodes_kwargs)} not list.'
    assert all(
        isinstance(node_args, expr) for node_args in nodes_args), (
        f'{repr(nodes_args)} not list of AST expression nodes.')
    assert all(
        isinstance(node_kwargs, keyword) for node_kwargs in nodes_kwargs), (
        f'{repr(nodes_kwargs)} not list of AST keyword nodes.')

    # Child node referencing the callable to be called.
    node_func_name = make_node_name_load(
        name=func_name, node_sibling=node_sibling)

    # Child node calling this callable.
    node_func_call = Call(
        func=node_func_name,
        args=nodes_args,
        keywords=nodes_kwargs,
    )

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_func_call)

    # Return this call node.
    return node_func_call

# ....................{ FACTORIES ~ call : arg             }....................
#FIXME: Unit test us up, please.
def make_node_kwarg(
    kwarg_name: str, kwarg_value: expr, node_sibling: AST) -> keyword:
    '''
    Create and return a new **keyword argument abstract syntax tree (AST) node**
    (i.e., node encapsulating a keyword argument of a call to an arbitrary
    function or method) passing the keyword argument with the passed name and
    value to some parent node encapsulating a call to some function or method.

    Parameters
    ----------
    kwarg_name : str
        Name of this keyword argument.
    kwarg_value : expr
        Expression node passing the value of this keyword argument.
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    -------
    keyword
        Keyword node passing a keyword argument with this name and value.
    '''
    assert isinstance(kwarg_name, str), f'{repr(kwarg_name)} not string.'
    assert isinstance(kwarg_value, expr), (
        f'{repr(kwarg_value)} not AST expression node.')

    # Child node encapsulating this keyword argument.
    node_kwarg = keyword(arg=kwarg_name, value=kwarg_value)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_kwarg)

    # Return this expression node.
    return node_kwarg

# ....................{ FACTORIES ~ import                 }....................
#FIXME: Unit test us up, please.
def make_node_importfrom(
    # Mandatory parameters.
    module_name: str,
    source_attr_name: str,
    node_sibling: AST,

    # Optional parameters.
    target_attr_name: Optional[str] = None,
) -> ImportFrom:
    '''
    Create and return a new **import-from abstract syntax tree (AST) node**
    (i.e., node encapsulating an import statement of the alias-style format
    ``from {module_name} import {attr_name}``) importing the attribute with the
    passed source name from the module with the passed name into the currently
    visited module as a new attribute with the passed target name.

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to import this attribute from.
    source_attr_name : str
        Unqualified basename of the attribute to import from this module.
    target_attr_name : Optional[str]
        Either:

        * If this attribute is to be imported into the currently visited module
          under a different unqualified basename, that basename.
        * If this attribute is to be imported into the currently visited module
          under the same unqualified basename as ``source_attr_name``,
          :data:`None`.

        Defaults to :data:`None`.
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    -------
    ImportFrom
        Import-from node importing this attribute from this module.
    '''
    assert isinstance(module_name, str), f'{repr(module_name)} not string.'
    assert isinstance(source_attr_name, str), (
        f'{repr(source_attr_name)} not string.')
    assert isinstance(target_attr_name, NoneTypeOr[str]), (
        f'{repr(target_attr_name)} neither string nor "None".')

    # Node encapsulating the name of the attribute to import from this module,
    # defined as either...
    node_importfrom_name = (
        # If this attribute is to be imported into the currently visited module
        # under a different basename, do so;
        alias(name=source_attr_name, asname=target_attr_name)
        if target_attr_name else
        # Else, this attribute is to be imported into the currently visited
        # module under the same basename. In this case, do so.
        alias(name=source_attr_name)
    )

    # Node encapsulating the name of the module to import this attribute from.
    node_importfrom = ImportFrom(
        module=module_name,
        names=[node_importfrom_name],
        # Force an absolute import for safety (i.e., prohibit relative imports).
        level=0,
    )

    # Copy all source code metadata (e.g., line numbers) from this sibling node
    # onto these new nodes.
    copy_node_metadata(
        node_src=node_sibling, node_trg=(node_importfrom, node_importfrom_name))

    # Return this import-from node.
    return node_importfrom

# ....................{ FACTORIES ~ literal : string       }....................
#FIXME: Unit test us up, please.
def make_node_str(text: str, node_sibling: AST) -> Constant:
    '''
    Create and return a new **string literal abstract syntax tree
    (AST) node** (i.e., node encapsulating the passed string).

    Parameters
    ----------
    text : str
        String literal to be encapsulated in a new node.
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    -------
    Constant
        String literal node encapsulating this string.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Child node encapsulating this string.
    node_str = Constant(value=text)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_str)

    # Return this string literal node.
    return node_str

# ....................{ FACTORIES ~ literal : f-string     }....................
#FIXME: Unit test us up, please.
def make_node_fstr_field(node_expr: expr, node_sibling: AST) -> FormattedValue:
    '''
    Create and return a new **f-string formatting field abstract syntax tree
    (AST) node** (i.e., node embedding the substring created and returned by the
    evaluation of the passed arbitrary expression in some parent node
    encapsulating an f-string embedding this field).

    This factory function creates substrings resembling ``{some_fstr_field}`` in
    larger f-strings resembling ``f'This is {some_fstr_field}, isn't it?'``.

    Caveats
    -------
    This field assumes *no* suffixing ``!``-prefixed conversion (e.g., "!a",
    "!r", "!s"). Thankfully, those conversions are only syntactic sugar for more
    human-readable builtins (e.g., ``repr()``, ``str()``). Ergo, this caveat
    does *not* actually constitute a hard constraint. Just prefer the builtins.

    Parameters
    ----------
    node_expr : expr
        Formatting field to be embedded in some parent f-string node.
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    -------
    Name
        Name node accessing this attribute in the current lexical scope.
    '''
    assert isinstance(node_expr, expr), (
        f'{repr(node_expr)} not AST expression node.')

    # Child node encapsulating a formatting field "{node_expr.value}" in some
    # parent node encapsulating an f-string embedding this field. For unknown
    # reasons, the standard "ast" module requires that the "conversion"
    # parameter be passed as a non-standard magic integer constant. Whatevahs!
    node_fstr_field = FormattedValue(value=node_expr, conversion=-1)

    # Copy source code metadata from this sibling node onto this new node.
    copy_node_metadata(node_src=node_sibling, node_trg=node_fstr_field)

    # Return this f-string field node.
    return node_fstr_field



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646` **data submodule.**

This submodule exercises :pep:`646` support for statement-level tuple and type
variable unpacking (e.g., ``*TypeVarTuple('Ts')``, ``*Tuple[int, str]``)
implemented in the :func:`beartype.beartype` decorator by declaring unit tests
exercising these aliases. For safety, these tests are intentionally isolated
from the main test suite. Notably, this low-level submodule implements the
higher-level ``test_decor_pep646()`` unit test in the main test suite.

Caveats
-------
**This submodule requires the active Python interpreter to target at least
Python 3.11.0.** If this is *not* the case, importing this submodule raises an
:exc:`SyntaxError` exception.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import TypeVarTuple
from beartype._util.hint.pep.proposal.pep646692 import (
    make_hint_pep646_tuple_unpacked_prefix,
    make_hint_pep646_typevartuple_unpacked_prefix,
    make_hint_pep646_typevartuple_unpacked_subbed,
)

# ....................{ HINTS                              }....................
TUPLE_UNPACKED_EMPTY = make_hint_pep646_tuple_unpacked_prefix(())
'''
:pep:`646`-compliant unpacked child tuple hint subscripted by the empty tuple
(signifying zero child hints).
'''

# ....................{ TYPEARGS                           }....................
Ts = TypeVarTuple('Ts')
'''
Arbitrary :pep:`646`-compliant type variable tuple.
'''


Us = TypeVarTuple('Us')
'''
Arbitrary :pep:`646`-compliant type variable tuple.
'''

# ....................{ TYPEARGS ~ unpacked                }....................
Ts_unpacked = Ts_unpacked_prefix = (
    make_hint_pep646_typevartuple_unpacked_prefix(Ts))
'''
Arbitrary :pep:`646`-compliant unpacked type variable tuple of the form ``*Ts``.
'''


Ts_unpacked_subbed = make_hint_pep646_typevartuple_unpacked_subbed(Ts)
'''
Arbitrary :pep:`646`-compliant unpacked type variable tuple of the form
``typing.Unpack[Ts]``.
'''


Us_unpacked = make_hint_pep646_typevartuple_unpacked_prefix(Us)
'''
Arbitrary :pep:`646`-compliant unpacked type variable tuple of the form ``*Ts``.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook integration test fixtures** (i.e., :mod:`pytest`-specific
context managers passed as parameters to integration tests exercising the
:mod:`beartype.claw` subpackage).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import fixture

# ....................{ FIXTURES ~ equality                }....................
@fixture(autouse=True, scope='function')
def clean_claws_func() -> None:  # <-- heh. get it... clean *CLAWS*? it's punny.
    '''
    Permanently, silently, and recursively remove all **bytecode files** (i.e.,
    pure-Python bytecode compiled to platform-dependent temporary files residing
    in temporary ``__pycache__/`` subdirectories) of the test-specific
    :mod:`beartype_test.a90_func.data` subpackage and all subsubpackages of that
    subpackage regardless of depth.

    Note that this integration test-scoped fixture is implicitly performed
    *before* each integration test transitively defined in sibling and child
    submodules of the subpackage directly containing this submodule. Why?
    Because failing to do so would invite subtle but easily reproducible
    desynchronization woes between those files and more recent changes to the
    implementation of the :mod:`beartype.claw` subpackage in the main codebase.

    See Also
    --------
    :func:`beartype._util.path.utilpathremove.remove_package_bytecode_files`
        Further details.
    '''

    # Defer fixture-specific imports.
    #
    # Note that we intentionally avoid importing any subsubpackages (e.g.,
    # "beartype_test.a00_unit.data.claw.intraprocess.hookable_package.beartype_this_package")
    # above. Why? Because doing so would implicitly install the exact beartype
    # import hook which calling unit tests are attempting to subsequently
    # exercise and which *MUST* be confined to a context manager for test
    # idempotency.
    from beartype_test.a90_func.data import claw
    from beartype_test._util.data.pytdataclean import (
        data_claw_subpackage_cleaned)

    # Recursively remove *ALL* previously compiled bytecode files from both this
    # subpackage *and* *ALL* subsubpackages of this subpackage.
    yield from data_claw_subpackage_cleaned(claw)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hookable Celery integration submodule** (i.e., data module
defining Celery-specific Runnables decorated by the third-party
decorator-hostile :func:`celery_core.runnables.chain` decorator which the
:mod:`beartype.beartype` decorator will then be injected *after* rather than
*before* as the earliest decorator, mimicking real-world usage of
:func:`beartype.claw` import hooks from external callers).

See Also
--------
https://github.com/beartype/beartype/issues/500
    StackOverflow issue exercised by this submodule.
'''

# ....................{ IMPORTS                            }....................
# from beartype.roar import BeartypeCallHintParamViolation
from celery import Celery
# from pytest import raises

# ....................{ GLOBALS                            }....................
celery_server = Celery(broker='memory://')
'''
Arbitrary in-memory Celery server leveraging the in-memory broker.

Note that Celery provides *no* equivalent in-memory backend, for the obvious
reason that Celery principally leverages the multiprocessing forking model.
Clearly, forks do *not* share memory. Unfortunately, because Celery provides
*no* equivalent in-memory backend, there exists *no* trivial means of extracting
the data asynchronously returned from a Celery task. The Celery task defined
below thus intentionally returns *no* data; instead, that task simply validates
the contents of the passed input.
'''

# ....................{ FUNCTIONS                          }....................
@celery_server.task(name="Amaz'd and full of fear; like anxious men")
def amazed_and_full_of_fear(like_anxious_men: dict[str, str]) -> None:
    '''
    Arbitrary function trivially satisfying Celery's task protocol by
    accepting an arbitrary dictionary of mock input data.

    The :obj:`Celery.task` decorator method wraps this function in a
    Celery-specific ``Task`` object defining an ``apply_async()`` method, which
    we then indirectly call below via the higher-level ``delay()`` method to
    validate this wrapping.

    Note that this Celery task intentionally returns *no* output. Doing so would
    require the non-trivial installation, configuration, and registration of a
    non-trivial third-party package above as this Celery server's backend.
    '''

    # Value associated with this key of this caller-defined mock input.
    gather_in_panting_troops = like_anxious_men['gather in panting troops']

    # Assert this value to be the expected mock input.
    assert gather_in_panting_troops == 'like anxious men'

# ....................{ PASS                               }....................
# Implicitly assert that the import hook registered by the caller respected the
# decorator-hostile @task decorator decorating the function defined above by
# injecting the @beartype decorator as the last decorator on that function.
#
# If this call raises an unexpected exception, then that import hook failed to
# respect that the decorator-hostile @task decorator by instead injecting the
# @beartype decorator as the first decorator on that function. Since @task is
# decorator-hostile and thus hostile to @beartype as well, @task prohibits
# *ANY* decorator from being applied after itself.
#
# Note that this result is extremely unsafe and thus largely useless. Since this
# Celery server lacks a backend, Celery tasks queued against this server can
# only be invoked. They *CANNOT* be queried after invocation -- even as to their
# failure state. Attempting to call *ANY* method of this result causes Celery to
# raise an unreadable exception resembling:
#     NotImplementedError: No result backend is configured.
#     Please see the documentation for more information.
#
# In other words, Celery is fundamentally hostile to testing by unrelated
# third-party packages like @beartype that could honestly care less about
# Celery. Seriously, Celery. What were you thinking?
amazed_and_full_of_fear_result = amazed_and_full_of_fear.delay(
    {'gather in panting troops': 'like anxious men'})

#FIXME: Preserved for posterity, but unlikely we'll ever be able to resurrect
#this. Doing so raises the "NotImplementedError" exception detailed above.
# # Value returned by that call.
# when_earthquakes_jar = amazed_and_full_of_fear_result.get()
#
# # Explicitly assert that call returned the expected output.
# print(f'when_earthquakes_jar: {when_earthquakes_jar}')
# assert when_earthquakes_jar == 'Who on wide plains like anxious men'

# ....................{ FAIL                               }....................
#FIXME: *LOLBRO.* We have no idea how to even test whether a Celery task raises
#an exception or not without attaching a valid third-party Celery backend to the
#Celery server defined above, which we *CLEARLY* are not going to do. *sigh*
# # Assert this Runnable when passed an invalid parameter raises the expected
# # @beartype-specific type-checking violation.
# with raises(BeartypeCallHintParamViolation):
#     amazed_and_full_of_fear.delay(
#         {len('When earthquakes jar'): len('their battlements and towers.')})



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **cache clearerers** (i.e., low-level callables safely resetting
global caches distributed throughout the :mod:`beartype` codebase).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ CLEARERS                           }....................
def clear_caches() -> None:
    '''
    Clear (i.e., empty) *all* internal caches leveraged throughout the
    :mod:`beartype` codebase, enabling callers to reset this codebase to its
    initial state.

    This function is typically cleared on detecting a **hot reload** (i.e.,
    attempt by the end user to reimport a presumably redefined user-defined
    module, type, or other object commonly cached by :mod:`beartype`). Notably,
    this function clears:

    * The **annotations dictionary cache** (i.e., private
      :data:`beartype._util.hint.pep.proposal.pep649._MODULE_NAME_TO_HINTABLE_BASENAME_TO_ANNOTATIONS`
      dictionary).
    * The **forward reference proxy cache** (i.e., private
      :data:`beartype._check.forward.reference.fwdrefmake._forwardref_args_to_forwardref`
      dictionary).
    * The **forward reference referee cache** (i.e., private
      :data:`beartype._check.forward.reference.fwdrefmeta._forwardref_to_referent`
      dictionary).
    * The **tuple union cache** (i.e., private
      :data:`beartype._check.code.codescope._tuple_union_to_tuple_union`
      dictionary).
    * The **type hint coercion cache** (i.e., private
      :data:`beartype._check.convert._convcoerce._hint_repr_to_hint`
      dictionary).
    * The **type hint wrapper cache** (i.e., private
      :data:`beartype._door._cls.doormeta._HINT_KEY_TO_WRAPPER` cache).
    '''
    # print('Clearing all \"beartype._check\" caches...')

    # Defer possibly heavyweight imports. Whereas importing this submodule is a
    # common occurrence, cache clearing and thus calls to this function are a
    # comparatively rarer occurrence. We optimize for the common case.
    from beartype.door._cls.doormeta import _HINT_KEY_TO_WRAPPER
    from beartype._check.code.codescope import _tuple_union_to_tuple_union
    from beartype._check.convert._convcoerce import _hint_repr_to_hint
    from beartype._check.forward.reference.fwdrefmake import (
        _forwardref_args_to_forwardref)
    from beartype._check.forward.reference.fwdrefmeta import (
        _forwardref_to_referent)
    from beartype._util.cache.utilcacheobjattr import clear_object_attr_caches

    # Clear all relevant caches used throughout this subpackage.
    clear_object_attr_caches()
    _HINT_KEY_TO_WRAPPER.clear()
    _forwardref_args_to_forwardref.clear()
    _forwardref_to_referent.clear()
    _hint_repr_to_hint.clear()
    _tuple_union_to_tuple_union.clear()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **string getters** (i.e., low-level callables slicing and returning
substrings out of arbitrary strings, typically to acquire prefixes and suffixes
satisfying various conditions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
# from beartype.roar._roarexc import _BeartypeUtilTextException

# ....................{ GETTERS                            }....................
#FIXME: Uncomment if ever needed.
# def get_str_prefix_greedy(text: str, anchor: str) -> str:
#     '''
#     **Greedily anchored prefix** (i.e., substring ranging from the first
#     character to the last instance of the passed substring) of the passed
#     string if any *or* raise an exception otherwise (i.e., if this string
#     contains no such substring).
#
#     Parameters
#     ----------
#     text : str
#         String to be searched.
#     anchor: str
#         Substring to search this string for.
#
#     Returns
#     ----------
#     str
#         Prefix of this string preceding the last instance of this substring.
#
#     Raises
#     ----------
#     _BeartypeUtilTextException
#         If this string contains *no* instance of this substring.
#
#     See Also
#     ----------
#     :func:`get_str_prefix_greedy_or_none`
#         Further details.
#     '''
#
#     # Greedily anchored prefix of this string if any *OR* "None" otherwise.
#     text_prefix_greedy = get_str_prefix_greedy_or_none(text, anchor)
#
#     # If this string contains *NO* such prefix, raise an exception.
#     if text_prefix_greedy is None:
#         raise _BeartypeUtilTextException(
#             f'String "{text}" substring "{anchor}" not found.')
#
#     # Else, return this prefix.
#     return text_prefix_greedy


#FIXME: Uncomment if ever needed.
# def get_str_prefix_greedy_or_none(text: str, anchor: str) -> 'Optional[str]':
#     '''
#     **Greedily anchored prefix** (i.e., substring ranging from the first
#     character to the last instance of the passed substring) of the passed
#     string if any *or* ``None`` otherwise (i.e., if this string contains no
#     such substring).
#
#     Parameters
#     ----------
#     text : str
#         String to be searched.
#     anchor: str
#         Substring to search this string for.
#
#     Returns
#     ----------
#     Optional[str]
#         Either:
#
#         * If this string contains this substring, the prefix of this string
#           preceding the last instance of this substring.
#         * Else, ``None``.
#
#     Examples
#     ----------
#         >>> from beartype._util.text.utiltextget import (
#         ...     get_str_prefix_greedy_or_none)
#         >>> get_str_prefix_greedy_or_none(
#         ...     text='Opposition...contradiction...premonition...compromise.',
#         ...     anchor='.')
#         Opposition...contradiction...premonition...compromise
#         >>> get_str_prefix_greedy_or_none(
#         ...     text='This is an anomaly. Disabled. What is true?',
#         ...     anchor='!')
#         None
#     '''
#     assert isinstance(text, str), f'{repr(text)} not string.'
#     assert isinstance(anchor, str), f'{repr(anchor)} not string.'
#
#     # Return either...
#     return (
#         # If this string contains this substring, the substring of this string
#         # preceding the last instance of this substring in this string.
#         text[:text.rindex(anchor)]
#         if anchor in text else
#         # Else, "None".
#         None
#     )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **beartype import hookable decorator-hostile decorator function
submodule** (i.e., defining callables decorated by chains of one or more
**decorator-hostile decorator functions** (i.e., decorators implemented as
functions but hostile to other decorators by prohibiting other decorators from
being applied after they themselves are applied in such a chain) into which the
:mod:`beartype.beartype` decorator will be injected after the last
decorator-hostile decorator, mimicking real-world usage of the
:func:`beartype.claw.beartype_package` import hook from external callers).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintReturnViolation
from beartype.typing import no_type_check
from functools import lru_cache
from pytest import raises

# ....................{ IMPORTS                            }....................
# Exercise all possible kinds of imports of this decorator-hostile decorator
# function.

# Import this decorator-hostile decorator function directly as an attribute from
# the nested submodule defining this function decorator via an absolute
# import-from statement.
from beartype_test.a00_unit.data.func.data_decor import (
    DecoratorHostileWrapper,
    decorator_hostile,

    # Import this same function under an aliased name.
    decorator_hostile as decorator_hostile_alias,
)

# Import the root package transitively defining this function with an absolute
# import statement.
import beartype_test

# Import the nested submodule defining this decorator with an absolute
# import-from statement.
from beartype_test.a00_unit.data.func import data_decor

# Import a nested submodule transitively defining this function with a relative
# import-from statement.
from ..... import func

# ....................{ FUNCTIONS ~ checkable              }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator *AFTER* the first decorator-hostile decorator
# function in all chains of callable decorations.

@decorator_hostile
def roused_from_icy_trance() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator function directly imported as an attribute from
    the submodule defining that function.
    '''

    return "Even now, while Saturn, rous'd from icy trance,"


@decorator_hostile
@decorator_hostile_alias
def his_palace_door_flew_ope() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by:

    * A decorator-hostile decorator function directly imported as an attribute
      from the submodule defining that function aliased to a different local
      name.
    * One or more other decorator-hostile decorator functions, exercising that
      :func:`beartype.claw.beartype_package` import hooks correctly handle
      chains of multiple decorator-hostile decorators.
    '''

    return 'Then, as was wont, his palace-door flew ope'


@beartype_test.a00_unit.data.func.data_decor.decorator_hostile
def thea_through_the_woods() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator function accessed via the root package
    transitively defining that function.
    '''

    return 'Went step for step with Thea through the woods,'


@data_decor.decorator_hostile
def twilight_in_the_rear() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator function accessed via the nested submodule
    directly defining that function.
    '''

    return 'Hyperion, leaving twilight in the rear,'


@func.data_decor.decorator_hostile
def threshold_of_the_west() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator function accessed via a nested submodule only
    transitively defining that function.
    '''

    return 'Came slope upon the threshold of the west;'

# ....................{ FUNCTIONS ~ uncheckable            }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator *AFTER* the first decorator-hostile decorator
# function but *BEFORE* the last decorator-friendly decorator function in all
# chains of callable decorations such that one of those decorator-friendly
# decorators is @typing.no_type_check, thus instructing the injected @beartype
# decorator to silently reduce to a noop (i.e., type-check *NOTHING*).

@decorator_hostile
@decorator_hostile_alias
@lru_cache
@no_type_check
def and_wandering_sounds() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by:

    * A decorator-hostile decorator function directly imported as an attribute
      from the submodule defining that function aliased to a different local
      name.
    * One or more other decorator-hostile decorator functions, exercising that
      :func:`beartype.claw.beartype_package` import hooks correctly handle
      chains of multiple decorator-hostile decorators.
    * One or more other decorator-friendly decorator functions, exercising that
      :func:`beartype.claw.beartype_package` import hooks correctly handle
      chains of multiple decorator-friendly decorators.
    '''

    return 'And wandering sounds, slow-breathed melodies;'

# ....................{ LOCALS                             }....................
# Tuple of all decorator-hostile decorator-decorated callables defined above.
FUNCS_DECOR_HOSTILE_CHECKED = (
    roused_from_icy_trance,
    his_palace_door_flew_ope,
    thea_through_the_woods,
    twilight_in_the_rear,
    threshold_of_the_west,
)

# ....................{ ASSERTS                            }....................
# For each decorator-hostile decorator-decorated callable defined above...
for func_decor_hostile_checked in FUNCS_DECOR_HOSTILE_CHECKED:
    # ....................{ PASS                           }....................
    # Assert that the import hook registered by the caller preserved the
    # uncallable object created and returned by the decorator-hostile decorator
    # decorating the callables defined above.
    assert isinstance(func_decor_hostile_checked, DecoratorHostileWrapper)

    # ....................{ FAIL                           }....................
    # Assert that the import hook registered by the caller injected the
    # @beartype decorator *AFTER* the last decorator-hostile decorator
    # decorating the callables defined above.
    with raises(BeartypeCallHintReturnViolation):
        func_decor_hostile_checked.invoke()

# ....................{ PASS                               }....................
# Assert that the import hook registered by the caller injected the @beartype
# decorator *AFTER* the last decorator-hostile decorator but *BEFORE* the last
# decorator-friendly decorator decorating the callable defined above, resulting
# in @beartype *NOT* type-checking the resulting uncallable object.
assert and_wandering_sounds.invoke() == (
    'And wandering sounds, slow-breathed melodies;')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Call stack utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.utilfunc.utilfuncframe` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getters                    }....................
def test_get_frame_or_none() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncframe.get_frame_or_none` getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallFrameException
    from beartype._util.func.utilfunccodeobj import get_func_codeobj
    from beartype._util.func.utilfuncframe import get_frame_or_none
    from pytest import raises
    from sys import maxsize

    # ....................{ CALLABLES                      }....................
    def while_fate() -> None:
        '''
        Arbitrary callable calling another callable calling another callable.
        '''

        seemd_strangled()


    def seemd_strangled() -> None:
        '''
        Arbitrary callable calling another callable.
        '''

        in_my_nervous_grasp()


    def in_my_nervous_grasp() -> None:
        '''
        Arbitrary callable calling *no* other callable but instead validating
        the stack frames on the current call stack are as expected.
        '''

        # ....................{ PASS                       }....................
        # Assert this getter returns the parent stack frame when passed *NO*
        # parameters.
        seemd_strangled_frame = get_frame_or_none()
        assert (
            get_func_codeobj(seemd_strangled_frame) is
            get_func_codeobj(seemd_strangled)
        )

        # Assert this getter returns the current stack frame when passed 0.
        in_my_nervous_grasp_frame = get_frame_or_none(0)
        assert (
            get_func_codeobj(in_my_nervous_grasp_frame) is
            get_func_codeobj(in_my_nervous_grasp)
        )

        # Assert this getter returns the parent parent stack frame when passed
        # 2.
        while_fate_frame = get_frame_or_none(2)
        assert (
            get_func_codeobj(while_fate_frame) is
            get_func_codeobj(while_fate)
        )

    # ....................{ PASS                           }....................
    # Call an arbitrary callable calling another callable calling another
    # callable validating the stack frames on the current call stack.
    while_fate()

    # Assert this getter returns "None" when the passed number of stack frames
    # to ignore is an absurdly large positive integer guaranteed to either:
    # * Trigger a C-based integer overflow.
    # * *NOT* trigger such an overflow but still exceed the total number of
    #   stack frames on the current call stack.
    assert get_frame_or_none(maxsize) is None  # <-- overflow!
    assert get_frame_or_none(9999) is None  # <-- no overflow!

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when the passed number of
    # stack frames to ignore is *NOT* a non-negative integer.
    with raises(_BeartypeUtilCallFrameException):
        get_frame_or_none("While Fate seem'd strangled in my nervous grasp?")
    with raises(_BeartypeUtilCallFrameException):
        get_frame_or_none(-1)


def test_get_frame() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncframe.get_frame` getter.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfuncframe import get_frame

    # Assert this attribute is a callable under both CPython and PyPy.
    assert callable(get_frame) is True

# ....................{ TESTS ~ iterators                  }....................
def test_iter_frames() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncframe.iter_frames` iterator.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunccodeobj import get_func_codeobj
    from beartype._util.func.utilfuncframe import iter_frames

    # For each stack frame on the current call stack iterated by this
    # iterator...
    for frame in iter_frames():
        # Code object underlying the pure-Python function encapsulated by this
        # stack frame.
        frame_codeobj = get_func_codeobj(frame)

        # Unqualified name of this function.
        frame_name = frame_codeobj.co_name

        # Assert that this function is the current unit test function.
        assert frame_name == 'test_iter_frames'

        # Halt iteration. For safety, avoid attempting to assert *ANYTHING*
        # about pytest-specific stack frames deeper in the call stack.
        break
    # If the prior "break" statement was *NOT* run, the prior iteration did
    # *NOT* run at all. In this case, raise an unconditional test failure.
    else:
        assert False, 'iter_frames() empty.'

    # Assert that this iterator reduces to the empty generator when asked to
    # ignore more stack frames than currently reside on the call stack.
    #
    # Note that an excessively large number of stack frames is intentionally
    # chosen to avoid conflicts with "pytest" itself, which is implemented in
    # pure-Python and thus runs tests with Python functions that themselves add
    # an arbitrary number of stack frames to the call stack.
    assert tuple(iter_frames(func_stack_frames_ignore=999999)) == ()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype validator text utilities** (i.e., callables performing low-level
string-centric operations on behalf of higher-level beartype validators).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeValeUtilException
from beartype.typing import Optional
from beartype._cave._cavemap import NoneTypeOr

# ....................{ FORMATTERS                         }....................
def format_diagnosis_line(
    # Mandatory parameters.
    validator_repr: str,
    indent_level_outer: str,
    indent_level_inner: str,

    # Optional parameters.
    is_obj_valid: Optional[bool] = None,
) -> str:
    '''
    Single line of a larger human-readable **validation failure diagnosis**
    (i.e., substring describing how an arbitrary object either satisfies *or*
    violates an arbitrary validator), formatted with the passed indentation
    level and boolean value.

    Parameters
    ----------
    validator_repr : str
        **Validator representation** (i.e., unformatted single line of a larger
        diagnosis report to be formatted by this function).
    indent_level_outer : str
        **Outermost indentation level** (i.e., zero or more adjacent spaces
        prefixing each line of the returned substring).
    indent_level_inner : str
        **Innermost indentation level** (i.e., zero or more adjacent spaces
        delimiting the human-readable representation of the tri-state boolean
        and validator representation in the returned substring).
    is_obj_valid : Optional[bool]
        Tri-state boolean such that:

        * If ``True``, that arbitrary object satisfies the beartype validator
          described by this specific line.
        * If ``False``, that arbitrary object violates the beartype validator
          described by this specific line.
        * If ``None``, this specific line is entirely syntactic (e.g., a
          suffixing ")" delimiter) isolated to its own discrete line for
          readability. In this case, this line does *not* describe how an
          arbitrary object either satisfies *or* violates an arbitrary
          validator.

        Defaults to ``None``.

    Returns
    ----------
    str
        This diagnosis line formatted with this indentation level.

    Raises
    ----------
    _BeartypeValeUtilException
        If ``is_obj_valid`` is *not* a **tri-state boolean** (i.e., either
        ``True``, ``False``, or ``None``).
    '''
    assert isinstance(validator_repr, str), (
        f'{repr(validator_repr)} not string.')
    assert isinstance(indent_level_outer, str), (
        f'{repr(indent_level_outer)} not string.')
    assert isinstance(indent_level_inner, str), (
        f'{repr(indent_level_inner)} not string.')

    # If "is_obj_valid" is *NOT* a tri-state boolean, raise an exception.
    #
    # Note that this condition is intentionally validated with full-blown
    # exception handling rather than a simple "assert" statement. This condition
    # was previously implemented via a simple "assert" statement, which then
    # raised a non-human-readable assertion in an end user issue. *OH, GODS!*
    if not isinstance(is_obj_valid, NoneTypeOr[bool]):
        raise _BeartypeValeUtilException(
            f'beartype.vale._valeutiltext.format_diagnosis_line() parameter '
            f'"is_obj_valid" value {repr(is_obj_valid)} '
            f'not tri-state boolean for '
            f'validator representation: {validator_repr}'
        )
    # Else, "is_obj_valid" is a tri-state boolean.

    # String representing this boolean value, padded with spaces on the left as
    # needed to produce a column-aligned line diagnosis resembling:
    #     False == (
    #      True ==     Is[lambda foo: foo.x + foo.y >= 0] &
    #     False ==     Is[lambda foo: foo.x + foo.y <= 10]
    #              )
    is_obj_valid_str = ''
    if   is_obj_valid is True:  is_obj_valid_str = ' True == '
    elif is_obj_valid is False: is_obj_valid_str = 'False == '
    else:                       is_obj_valid_str = '         '

    # Do one thing and do it well.
    return (
        f'{indent_level_outer}'
        f'{is_obj_valid_str}'
        f'{indent_level_inner}'
        f'{validator_repr}'
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.


'''
**Test configuration** (i.e., :mod:`pytest`-specific early-time configuration
guaranteed to be implicitly imported by :mod:`pytest` into *all* sibling and
child submodules of the test subpackage containing this :mod:`pytest` plugin).
'''

# ....................{ IMPORTS                            }....................
# Import all subpackage-specific fixtures implicitly required by tests defined
# by submodules of this subpackage.
from beartype_test.a90_func.a40_claw._clawfixture import clean_claws_func



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype utility fixed list pool unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.cache.pool.utilcachepoolinstance` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from io import StringIO
from pytest import raises

# ....................{ TESTS                              }....................
def test_utilcachepool_instance() -> None:
    '''
    Test successful usage of the
    :mod:`beartype._util.cache.pool.utilcachepoolinstance` submodule.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepoolinstance import (
        acquire_instance,
        release_instance,
    )

    # ....................{ CONSTANTS                      }....................
    # Culturally relevant Clash lyrics to be tested below.
    PUBLIC_SERVICE_ANNOUNCEMENT = '\n'.join((
        'You have the right not to be killed.',
        'Murder is a crime,',
        'Unless it was done',
        'By a policeman',
        'Or an aristocrat.',
    ))
    KNOW_YOUR_RIGHTS = '\n'.join((
        'You have the right to food money --',
        'Providing, of course, you',
        "Don't mind a little",
        'Investigation, humiliation,',
        'And (if you cross your fingers)',
        'Rehabilitation.',
    ))

    # ....................{ PASS ~ container               }....................
    # Repeat this test twice, thus asserting that builtin mutable containers are
    # necessarily cleared by the acquire_instance() function.
    for _ in range(2):
        # Acquire an arbitrary dictionary.
        look_up = acquire_instance(dict)

        # Assert this dictionary to be empty.
        assert not look_up

        # Add arbitrary key-value pairs to this dictionary.
        look_up.update({
            'Look up,': 'and let me see our doom in it;',
            'Look up, and': 'tell me if this feeble shape.',
        })

        # Release this dictionary.
        release_instance(look_up)

    # Repeat this test twice, thus asserting that builtin mutable containers are
    # necessarily cleared by the acquire_instance() function.
    for _ in range(2):
        # Acquire an arbitrary list.
        is_saturns = acquire_instance(list)

        # Assert this list to be empty.
        assert not is_saturns

        # Add arbitrary key-value pairs to this list.
        is_saturns.extend((
            "Is Saturn's;", 'tell me,', "if thou hear'st the voice",))

        # Release this list.
        release_instance(is_saturns)

    # Repeat this test twice, thus asserting that builtin mutable containers are
    # necessarily cleared by the acquire_instance() function.
    for _ in range(2):
        # Acquire an arbitrary set.
        of_saturn = acquire_instance(set)

        # Assert this set to be empty.
        assert not of_saturn

        # Add arbitrary key-value pairs to this set.
        of_saturn.update({
            'Of Saturn;', 'tell me,', 'if this wrinkling brow,',})

        # Release this set.
        release_instance(of_saturn)

    # ....................{ PASS ~ stringio                }....................
    # Acquire an arbitrary string buffer.
    public_service_announcement = acquire_instance(StringIO)

    # Clear this buffer and reset its position to the start.
    public_service_announcement.truncate(0)
    public_service_announcement.seek(0)

    # Write a series of culturally relevant Clash lyrics to this buffer.
    public_service_announcement.write('You have the right not to be killed.\n')
    public_service_announcement.write('Murder is a crime,\n')
    public_service_announcement.write('Unless it was done\n')
    public_service_announcement.write('By a policeman\n')
    public_service_announcement.write('Or an aristocrat.')

    # Acquire another arbitrary string buffer.
    know_your_rights = acquire_instance(StringIO)

    # Clear this buffer and reset its position to the start.
    know_your_rights.truncate(0)
    know_your_rights.seek(0)

    # Write another series of culturally relevant Clash lyrics to this buffer.
    know_your_rights.write('You have the right to food money --\n')
    know_your_rights.write('Providing, of course, you\n')
    know_your_rights.write("Don't mind a little\n")
    know_your_rights.write('Investigation, humiliation,\n')
    know_your_rights.write('And (if you cross your fingers)\n')
    know_your_rights.write('Rehabilitation.')

    # Assert the contents of these buffers to still be as expected.
    assert (
        public_service_announcement.getvalue() == PUBLIC_SERVICE_ANNOUNCEMENT)
    assert know_your_rights.getvalue() == KNOW_YOUR_RIGHTS

    # Release the first buffer back to its parent pool.
    release_instance(public_service_announcement)

    # Reacquire the same buffer again.
    public_service_announcement_too = acquire_instance(StringIO)

    # Assert this to be the same buffer.
    assert public_service_announcement is public_service_announcement_too

    # Assert the second buffer to *NOT* be the same buffer.
    assert public_service_announcement is not know_your_rights

    # Release these buffers back to their parent pools (in acquisition order).
    release_instance(public_service_announcement)
    release_instance(know_your_rights)

    # ....................{ FAIL                           }....................
    # Assert that typed objects may only be acquired with types.
    with raises(AssertionError):
        acquire_instance((
            'You have the right to free speech',
            'As long as',
            "You're not dumb enough to actually try it.",
        ))
    with raises(AssertionError):
        acquire_instance(1977)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide :pep:`646`-compliant **generics** (i.e., subclasses of some
combination of the standard :pep:`484`-compliant :class:`typing.Generic`
superclass *and/or* a :pep:`585`-compliant type hint factory subscripted by one
or more type parameters such that one or more of these type parameters is a
:pep:`646`-compliant unpacked type variable tuple, exercising edge cases in unit
tests requiring non-trivial generics).

Caveats
-------
Whereas most PEP-compliant data submodules define type hints unique to this test
suite, this :pep:`646`-compliant data submodule mostly defines type hints
intentionally copy-pasted from :pep:`646` itself. Why? Non-triviality.
:pep:`646` is non-trivial, encompassing such a wide variety of edge cases that
:pep:`646` itself is currently Python's longest :mod:`typing` standard and
possibly Python's longest standard period. Taming the uncommon beast that is
:pep:`646` requires a slavish attention to detail best achieved by explicitly
supporting all type hints exhibited in :pep:`646` itself.

As an aid to readability and maintainability, most commented headings of this
submodule are directly taken from section headings of :pep:`646` itself.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Generic

# ....................{ "Summary Examples"                 }....................
# PEP 646's leading "Summary Examples" section defines the core PEP
# 646-compliant "Array" generic subsequently referenced throughout PEP 646. All
# following code is copy-pasted directly from that section.
from typing import TypeVar, TypeVarTuple

DType = TypeVar('DType')
Shape = TypeVarTuple('Shape')

class Array(Generic[DType, *Shape]):
    '''
    :pep:`646`-compliant generic subclassing the standard :pep:`484`-compliant
    :class:`typing.Generic` superclass parametrized by a single
    :pep:`646`-compliant unpacked type variable tuple.

    Note that this class declaration slightly varies from that specified in PEP
    646. Type hints annotating methods self-referencing this class have
    necessarily been stringified to enable their evaluation if the active Python
    interpreter targets Python < 3.14.
    '''

    def __abs__(self) -> 'Array[DType, *Shape]': ...

    def __add__(self, other: 'Array[DType, *Shape]') -> 'Array[DType, *Shape]': ...



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`557`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep557Exception
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import TypeException
from beartype._data.hint.sign.datahintsigns import (
    HintSignPep557DataclassInitVar)

# ....................{ GETTERS                            }....................
def get_hint_pep557_initvar_arg(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep557Exception,
    exception_prefix: str = '',
) -> Hint:
    '''
    PEP-compliant child type hint subscripting the passed :pep:`557`-compliant
    **dataclass initialization-only instance variable type hint** (i.e.,
    subscription of the :class:`dataclasses.InitVar` type hint factory).

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep557Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Hint
        PEP-compliant child type hint subscripting this parent type hint.

    Raises
    ------
    BeartypeDecorHintPep557Exception
        If this object is *not* a dataclass initialization-only instance
        variable type hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # Sign uniquely identifying this hint if this hint is identifiable *OR*
    # "None" otherwise.
    hint_sign = get_hint_pep_sign_or_none(hint)

    # If this hint is *NOT* a dataclass initialization-only instance variable
    # type hint, raise an exception.
    if hint_sign is not HintSignPep557DataclassInitVar:
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not '
            f'PEP 557-compliant "dataclasses.TypeVar" instance.'
        )
    # Else, this hint is such a hint.

    # Return the child type hint subscripting this parent type hint. Yes, this
    # hint exposes this child via a non-standard instance variable rather than
    # the "__args__" dunder tuple standardized by PEP 484.
    return hint.type  # type: ignore[attr-defined]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484` and :pep:`585` **generic type hint tester** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest`
submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ testers                    }....................
def test_is_hint_pep484585_generic(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest.is_hint_pep484585_generic`
    tester.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._data.hint.sign.datahintsignset import HINT_SIGNS_GENERIC
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
        is_hint_pep484585_generic)
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP

    # Assert this tester:
    # * Accepts generic PEP 484-compliant generics.
    # * Rejects concrete PEP-compliant type hints.
    for hint_pep_meta in hints_pep_meta:
        assert is_hint_pep484585_generic(hint_pep_meta.hint) is (
            hint_pep_meta.pep_sign in HINT_SIGNS_GENERIC)

    # Assert this tester rejects non-PEP-compliant type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        assert is_hint_pep484585_generic(not_hint_pep) is False


def test_is_hint_pep484585_generic_unsubbed(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest.is_hint_pep484585_generic_unsubbed`
    tester.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._data.hint.sign.datahintsigns import (
        HintSignPep484585GenericUnsubbed)
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
        is_hint_pep484585_generic_unsubbed)
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP

    # Assert this tester:
    # * Accepts generic PEP 484-compliant generics.
    # * Rejects concrete PEP-compliant type hints.
    for hint_pep_meta in hints_pep_meta:
        # print(f'hint_pep_meta: {repr(hint_pep_meta)}')
        assert is_hint_pep484585_generic_unsubbed(hint_pep_meta.hint) is (
            hint_pep_meta.pep_sign is HintSignPep484585GenericUnsubbed)

    # Assert this tester rejects non-PEP-compliant type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        assert is_hint_pep484585_generic_unsubbed(not_hint_pep) is False


def test_is_hint_pep484585_generic_subbed(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest.is_hint_pep484585_generic_subbed`
    tester.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._data.hint.sign.datahintsigns import (
        HintSignPep484585GenericSubbed)
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
        is_hint_pep484585_generic_subbed)
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP

    # Assert this tester:
    # * Accepts generic PEP 484-compliant generics.
    # * Rejects concrete PEP-compliant type hints.
    for hint_pep_meta in hints_pep_meta:
        # print(f'hint_pep_meta: {repr(hint_pep_meta)}')
        assert is_hint_pep484585_generic_subbed(hint_pep_meta.hint) is (
            hint_pep_meta.pep_sign is HintSignPep484585GenericSubbed)

    # Assert this tester rejects non-PEP-compliant type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        assert is_hint_pep484585_generic_subbed(not_hint_pep) is False

# ....................{ TESTS ~ testers : user             }....................
def test_is_hint_pep484585_generic_user(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest.is_hint_pep484585_generic_user`
    tester.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
        is_hint_pep484585_generic,
        is_hint_pep484585_generic_user,
    )
    from beartype.typing import (
        Generic,
        Protocol,
        Sequence,
        TypeVar,
    )

    # ....................{ TYPEVARS                       }....................
    # Arbitrary unconstrained type variables referenced below.
    T = TypeVar('T')

    # ....................{ LOCALS                         }....................
    # Tuple of zero or more generics that are *NOT* user-defined by third-party
    # logic residing outside the standard Python library.
    GENERICS_NONUSER = (
        Generic,
        Generic[T],
        Protocol,
        Protocol[T],
        Sequence[T],
    )

    # ....................{ ASSERTS                        }....................
    # Assert this tester rejects all non-user-defined generics.
    for generic_nonuser in GENERICS_NONUSER:
        assert is_hint_pep484585_generic_user(generic_nonuser) is False

    # Assert this tester accepts all user-defined generics.
    for hint_pep_meta in hints_pep_meta:
        generic_user = hint_pep_meta.hint
        assert is_hint_pep484585_generic_user(generic_user) is (
               is_hint_pep484585_generic(generic_user))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`563` + :pep:`695` **integration data submodule.**

This submodule exercises edge cases when combining :pep:`563` via the standard
``from __future__ import annotations`` pragma with :pep:`695`-compliant type
parametrizations known to interact problematically with :pep:`563`.
'''

# ....................{ IMPORTS                            }....................
from __future__ import annotations
from beartype import beartype
from beartype.roar import  (
    BeartypeDecorHintForwardRefException,
    BeartypeDecorHintPep695Exception,
)
from beartype.typing import (
    Optional,
)
from pytest import raises

# ....................{ CALLABLES                          }....................
@beartype
def shook_horrid[T](with_such_aspen_malady: list[T]) -> T | None:
    '''
    Arbitrary callable decorated by :func:`beartype.beartype`, parametrized by a
    :pep:`695`-compliant type parameter, and accepting an arbitrary parameter
    annotated by that type parameter.

    Returns
    -------
    T | None
        Either:

        * If this list is non-empty, the first item of this list.
        * Else, :data:`None`.
    '''

    # Do not panic. Have no fear. @beartype is here!
    return with_such_aspen_malady[0] if with_such_aspen_malady else None

# ....................{ CLASSES                            }....................
PeersLike = 'Peers like the front of Saturn. Who had power'
'''
Arbitrary global attribute sharing the same name as that of a
:pep:`695`-compliant type parameter parametrizing a callable defined below and
thus implicitly shadowed by that type parameter.
'''


@beartype
class NakedAndBare[S, T]:
    '''
    Arbitrary class decorated by :func:`beartype.beartype` and parametrized by
    multiple :pep:`695`-compliant type parameters.
    '''

    def of_its_great_diadem[PeersLike](
        self, the_front_of_saturn: (
        list[PeersLike] | tuple[S] | set[T])) -> (
        Optional[PeersLike]):
        '''
        Arbitrary callable decorated by :func:`beartype.beartype` and
        parametrized by a :pep:`695`-compliant type parameter whose name is the
        same as that of a global attribute that is *not* a type parameter.
        '''

        # Have no fear. @beartype is here... yet again, we win!
        return next(iter(the_front_of_saturn)) if the_front_of_saturn else None

# ....................{ FAIL                               }....................
# Assert that parametrizing a function by a PEP 695-compliant type parameter
# annotated by a syntactically invalid type hint subscripted by that parameter
# raises the expected exception.
with raises(BeartypeDecorHintForwardRefException):
    # Since the builtin "bytearray" type is *NOT* subscriptable, the following
    # type hint is syntactically invalid and raises a "TypeError" exception when
    # the @beartype PEP 563 resolver dynamically resolves this stringified hint.
    @beartype
    def but_it_is_so[U](and_i_am_smotherd_up: bytearray[U]) -> U:
        '''
        Arbitrary callable decorated by :func:`beartype.beartype` and
        parametrized by a :pep:`695`-compliant type parameter annotated by a
        syntactically invalid type hint subscripted by that parameter.
        '''

        # One-liners know no shame. Neither does our exhaustion.
        return b"But it is so, and I am smother'd up,"


# Assert that parametrizing a method by a PEP 695-compliant type parameter whose
# name is the same as another type parameter parametrizing the class defining
# that method raises the expected exception. PEP 695 explicitly prohibits this,
# which it refers to as "type parameter reuse."
with raises(BeartypeDecorHintPep695Exception):
    @beartype
    class TheaIFeelThee[T]:
        '''
        Arbitrary class decorated by :func:`beartype.beartype` and parametrized
        by a :pep:`695`-compliant type parameter.
        '''

        def ere_i_see[T](self, thy_face: T) -> T:
            '''
            Arbitrary callable decorated by :func:`beartype.beartype` and
            parametrized by a :pep:`695`-compliant type parameter whose name is
            the same as that of the type parameter parametrizing this class.
            '''

            return thy_face



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype key pool unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.cache.pool.utilcachepool` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                             }....................
def test_key_pool_pass() -> None:
    '''
    Test successful usage of the
    :class:`beartype._util.cache.pool.utilcachepool.KeyPool` type.
    '''

    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepool import KeyPool
    from beartype.roar._roarexc import _BeartypeUtilCachedKeyPoolException
    from io import StringIO
    from pytest import raises

    # Key pool to be tested, seeding empty pools keyed on the "newline"
    # parameter passed to the StringIO.__init__() method with a new "StringIO"
    # instance initialized to that parameter.
    #
    # Note that the "initial_value" parameter is currently unavailable under
    # "pypy3" and *MUST* thus be omitted here:
    #     $ pypy3
    #     >>> from io import StringIO
    #     >>> StringIO(initial_value='', newline=newline)
    #     TypeError: __init__() got an unexpected keyword argument 'initial_value'
    key_pool = KeyPool(item_maker=lambda newline: StringIO(newline=newline))

    # Acquire a new "StringIO" buffer writing Windows-style newlines.
    windows_stringio = key_pool.acquire(key='\r\n', is_debug=True)

    # Stanzas delimited by Windows-style newlines to be tested below.
    THE_DEAD_SHALL_BE_RAISED_INCORRUPTIBLE = '\r\n'.join((
        'My stomach, which has digested',
        'four hundred treaties giving the Indians',
        'eternal right to their land, I give to the Indians,',
        'I throw in my lungs which have spent four hundred years',
        'sucking in good faith on peace pipes.',
    ))
    BOOK_OF_NIGHTMARES = '\r\n'.join((
        'To the last man surviving on earth',
        'I give my eyelids worn out by fear, to wear',
        'in his long nights of radiation and silence,',
        'so that his eyes can’t close, for regret',
        'is like tears seeping through closed eyelids.',
    ))

    # Write a series of POSIX-style lines to this buffer.
    windows_stringio.write('My stomach, which has digested\n')
    windows_stringio.write('four hundred treaties giving the Indians\n')
    windows_stringio.write('eternal right to their land, I give to the Indians,\n')
    windows_stringio.write('I throw in my lungs which have spent four hundred years\n')
    windows_stringio.write('sucking in good faith on peace pipes.')

    # Verify the buffer implicitly converted all POSIX- to Windows-style
    # newlines in the resulting string.
    assert (
        windows_stringio.getvalue() == THE_DEAD_SHALL_BE_RAISED_INCORRUPTIBLE)

    # Release this buffer back to its parent pool.
    key_pool.release(key='\r\n', item=windows_stringio, is_debug=True)

    # Reacquire the same buffer.
    windows_stringio_too = key_pool.acquire(key='\r\n', is_debug=True)

    # Confirm the release-require cycle returns the same object
    assert windows_stringio is windows_stringio_too

    # Acquire another new "StringIO" buffer writing Windows-style newlines.
    windows_stringio_new = key_pool.acquire(key='\r\n', is_debug=True)

    # Assert this to *NOT* be the same buffer.
    assert windows_stringio is not windows_stringio_new

    # Write a series of POSIX-style lines to this new buffer.
    windows_stringio_new.write('To the last man surviving on earth\n')
    windows_stringio_new.write('I give my eyelids worn out by fear, to wear\n')
    windows_stringio_new.write('in his long nights of radiation and silence,\n')
    windows_stringio_new.write('so that his eyes can’t close, for regret\n')
    windows_stringio_new.write('is like tears seeping through closed eyelids.')

    # Confirm the new buffer also implicitly converted all POSIX- to
    # Windows-style newlines in the resulting string.
    assert windows_stringio_new.getvalue() == BOOK_OF_NIGHTMARES

    # Release these buffers back to their parent pools (in acquisition order).
    key_pool.release(key='\r\n', item=windows_stringio, is_debug=True)
    key_pool.release(key='\r\n', item=windows_stringio_new, is_debug=True)

    # Confirm the above object is released AND that releasing an already
    # released object with debugging logic enabled elicits a roar.
    with raises(_BeartypeUtilCachedKeyPoolException):
        key_pool.release(key='\r\n', item=windows_stringio_new, is_debug=True)


def test_key_pool_fail() -> None:
    '''
    Test unsuccessful usage of the
    :class:`beartype._util.cache.pool.utilcachepool.KeyPool` type.
    '''

    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepool import KeyPool
    from beartype.roar._roarexc import _BeartypeUtilCachedKeyPoolException
    from pytest import raises

    # Key pool to be tested, seeding empty pools with the identity function.
    key_pool = KeyPool(item_maker=lambda key: key)

    # Verify using an unhashable key to acquire a new object throws a TypeError
    with raises(TypeError):
        key_pool.acquire(
            ['Lieutenant!', 'This corpse will not stop burning!'],
            is_debug=True,
        )

    # Verify releasing a non-existent object elicits a roar
    with raises(_BeartypeUtilCachedKeyPoolException):
        key_pool.release(key='I should roar', item=object(), is_debug=True)
